// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

#include "MAdElasticityOp.h"
#include "CPUTimeManager.h"
#include "OperatorTools.h"
#include "MathUtils.h"
#include "MAdOutput.h"
#include "CallBackManager.h"

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
using std::make_pair;
using std::map;
using std::set;

namespace MAd {

  // -------------------------------------------------------------------
  void ElasticityOpCBFunction (pPList before, pPList after, void * data,
                               operationType type , pEntity ppp) 
  {
    MAdElasticityOp * elast = (MAdElasticityOp *)(data);
    if ( !(elast->relocationsComputed()) ) return;

    switch (type) {
    case MAd_ESPLIT: {
      
      // find the old edge
      void * temp = NULL;
      pEntity pE = PList_next(before,&temp);

      // take the mean of the displacements of the extremities 
      // of the edge for the new vertex
      elast->addVertexOnEdge( (pVertex)ppp, (pEdge)pE );

      break;
    } 
    case MAd_ECOLLAPSE: {
      elast->removeVertex( (pVertex)ppp );
      break;
    }
    case MAd_FSWAP:
    case MAd_ESWAP: {
      break;
    }
    case MAd_RREMOVE: {
      void * temp = NULL;
      while ( pEntity pE = PList_next(before,&temp) ) {
        if ( EN_type(pE) == 0 ) {
          elast->removeVertex( (pVertex)pE );
        }
      }
      break;
    }
    default: {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Callback function not implemented for mesh modification %d",
                                  type);
    }
    }
  }

  // -------------------------------------------------------------------
  MAdElasticityOp::MAdElasticityOp(pMesh m):
    mesh(m), E(1.), nu(0.45), chi(1.0), computed(false),
    cavityEqualMesh(false), cavityThickness(3)
  {
    dim = M_dim(mesh);
    if (dim!=3) {
      cerr << "Error: elasticity model only implemented for dimension 3, current: " << dim << endl;
      throw;
    }
    CallBackManagerSgl::instance().registerCallBack(ElasticityOpCBFunction,this);
  }

  // -------------------------------------------------------------------
  MAdElasticityOp::MAdElasticityOp(const MAdElasticityOp& eop):
    mesh(eop.mesh), dim(eop.dim), 
    E(eop.E), nu(eop.nu), chi(eop.chi),
    computed(eop.computed),
    cavityEqualMesh(eop.cavityEqualMesh),
    cavityThickness(eop.cavityThickness)
  {
    localIds = eop.localIds;
    dirichlet = eop.dirichlet;
    CallBackManagerSgl::instance().registerCallBack(ElasticityOpCBFunction,this);
  }

  // -------------------------------------------------------------------
  MAdElasticityOp::~MAdElasticityOp()
  {
    CallBackManagerSgl::instance().unregisterCallBack(ElasticityOpCBFunction,this);
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::setMaterials(double _E, double _nu)
  {
    E = _E;  nu = _nu;
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::setStiffnessAlterationCoef(double _chi)
  {
    chi = _chi;
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::clear()
  {
    localIds.clear();
    cavity.clear();
    dirichlet.clear();
    relocations.clear();
    computed = false;
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::buildCavity()
  {
    computed = false;

    if ( cavityEqualMesh ) return;

    int nVertInit = dirichlet.size();

    cavity.clear();

    if ( cavityThickness < 1 ) return;

    set<pVertex> innerNodes; // nodes inside the cavity and dirichlet
    set<pRegion> prevLayer;

    // --- first layer ---
    pRegion pr;
    pVertex pv;
    map<pVertex,smallVector>::const_iterator dIter = dirichlet.begin();
    for (; dIter != dirichlet.end(); dIter++) {
      pv = (*dIter).first;
      pPList vRegs = V_regions(pv);
      void * temp = NULL;
      while ( ( pr = (pRegion)PList_next(vRegs,&temp) ) ) {
        cavity.insert(pr);
        prevLayer.insert(pr);
      }
      PList_delete(vRegs);
      innerNodes.insert(pv);
    }

    // --- next layers ---
    for (int iL=1; iL < cavityThickness; iL++) {
      set<pRegion> curLayer;
      set<pRegion>::const_iterator rIter = prevLayer.begin();
      for (; rIter != prevLayer.end(); rIter++ ) {
        pPList rVerts = R_vertices(*rIter);
        void * temp = NULL;
        while ( ( pv = (pVertex)PList_next(rVerts,&temp) ) ) {
          if ( innerNodes.find(pv) == innerNodes.end() ) {
            pPList vRegs = V_regions(pv);
            void * temp = NULL;
            while ( ( pr = (pRegion)PList_next(vRegs,&temp) ) ) {
              if ( cavity.find(pr) == cavity.end() ) {
                cavity.insert(pr);
                curLayer.insert(pr);
              }
            }
            PList_delete(vRegs);
            innerNodes.insert(pv);
          }
        }
        PList_delete(rVerts);
      }
      prevLayer.clear();
      prevLayer.insert(curLayer.begin(),curLayer.end());
      curLayer.clear();
    }

    printf("Built a cavity with %d elements, dirichlet size: %d\n",cavity.size(),dirichlet.size());

  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::setHomogDirichletBC()
  {
    dirichlet.clear();
    setDirichletBC();
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::setDirichletBC()
  {
    computed = false;
    smallVector zero(3);  zero.set_all(0.);

    set<pVertex> verts;
    collectBCVertices(&verts);

    set<pVertex>::const_iterator vIter = verts.begin();
    set<pVertex>::const_iterator vLast = verts.end();
    for (; vIter != vLast; vIter++) {
      addDirichlet(*vIter,zero);
    }
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::collectBCVertices(set<pVertex> * bcVerts)
  {
    if ( cavityEqualMesh ) 
      {
        FIter fit = M_faceIter(mesh);
        while ( pFace pf = FIter_next(fit) ) {
          if ( EN_whatInType((pEntity)pf) == 2 ) {
            pPList fV = F_vertices(pf,1);
            void* tmp=0;
            while ( pVertex pv = (pVertex)PList_next(fV,&tmp) ) {
              (*bcVerts).insert(pv);
            }
            PList_delete(fV);
          }
        }
        FIter_delete(fit);
      }
    else
      {
        set<pRegion>::const_iterator cavIter = cavity.begin();
        for (; cavIter != cavity.end(); cavIter++) {
          pPList rFaces = R_faces(*cavIter);
          void * temp = NULL;
          pFace pf;
          while ( ( pf = (pFace)PList_next(rFaces,&temp) ) ) {
            pRegion otherR = F_otherRegion(pf,*cavIter);
            if ( !otherR || cavity.find(otherR) == cavity.end() ) {
              pPList fVerts = F_vertices(pf,1);
              void * temp2 = NULL;
              pVertex pv;
              while ( ( pv = (pVertex)PList_next(fVerts,&temp2) ) ) {
                (*bcVerts).insert(pv);
              }
              PList_delete(fVerts);
            }
          }
          PList_delete(rFaces);
        }
      }
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::addDirichlet(pVertex pv, smallVector& dxyz)
  {
    map<pVertex,smallVector>::iterator dIter = dirichlet.find(pv);
    if ( dIter != dirichlet.end() ) {
      smallVector dxyz1 = (*dIter).second;
      dxyz1.add(dxyz);
      dirichlet.erase(dIter);
      dirichlet.insert(make_pair(pv,dxyz1));
    }
    else {
      dirichlet.insert(make_pair(pv,dxyz));
    }
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::delDirichlet(pVertex pv)
  {
    map<pVertex,smallVector>::iterator dIter = dirichlet.find(pv);
    if ( dIter != dirichlet.end() ) {
      dirichlet.erase(dIter);
    }
  }

  // -------------------------------------------------------------------
  // Generates table of vertex pointers to local ids and returns number
  // of vertices in the cavity.
  int MAdElasticityOp::generateVertexIds()
  {
    computed = false;
    localIds.clear();

    int id = 0;

    if (cavityEqualMesh) {
      VIter vit = M_vertexIter(mesh);
      while ( pVertex pv = VIter_next(vit) ) {
        localIds[pv] = id;
        id++;
      }
      VIter_delete(vit);
    }
    else {
      set<pVertex> verts;
      set<pRegion>::const_iterator cavIter = cavity.begin();
      for (; cavIter != cavity.end(); cavIter++) {
        pPList rVerts = R_vertices(*cavIter);
        void * temp = NULL;
        while ( pVertex pv = (pVertex)PList_next(rVerts,&temp) ) {
          if ( verts.find(pv) == verts.end() ) {
            localIds[pv] = id;
            id++;
            verts.insert(pv);
          }
        }
        PList_delete(rVerts);
      }
    }
    return id;
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::addToMatrix(const pRegion pr,
                                    int nbVert,
                                    MAdLinearSystemDef * lsys)
  {
    int rNodes = R_numVertices(pr);
    int locSyze = rNodes*dim;

    smallMatrix locK(locSyze,locSyze);
    element3DMatrix(pr, locK);

    pPList vertsI = R_vertices(pr);
    pPList vertsJ = R_vertices(pr);
    void* tmpI = 0;
    int i = 0;
    while ( pVertex pvI = (pVertex)PList_next(vertsI,&tmpI) ) {

      int locIdI = localIds[pvI];
      
      void* tmpJ = 0;
      int j = 0;
      while ( pVertex pvJ = (pVertex)PList_next(vertsJ,&tmpJ) ) {
        
        int locIdJ = localIds[pvJ];
        
        for (int k=0; k<dim; k++) {
          for (int l=0; l<dim; l++) {
            lsys->addToMatrix(nbVert*k + locIdI, nbVert*l+locIdJ, locK(k*rNodes+i,l*rNodes+j));
          }
        }
        j++;
      }
      i++;
    }
    PList_delete(vertsI);
    PList_delete(vertsJ);
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::allocateMatrix(int nbVert, MAdLinearSystemDef * lsys)
  {
    if (cavityEqualMesh) {
      VIter vit2 = M_vertexIter(mesh);
      while ( pVertex pv2 = VIter_next(vit2) ) {
        int locId = localIds[pv2];
        int nbNeighbors = V_numEdges(pv2);
        for (int iDir=0; iDir<3; iDir++) {
          int row = nbVert*iDir + locId;
          lsys->set_nnz(row,(nbNeighbors+1)*3);// A node is coupled with all its neighbors, and don't forget himself
        }
      }
      VIter_delete(vit2);
    }
    else {
      map<pVertex,set<pVertex> > conn;
      set<pRegion>::const_iterator cIter = cavity.begin();
      for (; cIter != cavity.end(); cIter++) {
        pPList rEdges = R_edges(*cIter);
        void * temp = NULL;
        pEdge edge;
        pVertex v0, v1;
        while ( ( edge = (pEdge)PList_next(rEdges,&temp)) ) {
          v0 = E_vertex(edge,0);
          v1 = E_vertex(edge,1);
          conn[v0].insert(v1);
          conn[v1].insert(v0);
        }
        PList_delete(rEdges);
      }
        
      pVertex pv;
      map<pVertex,set<pVertex> >::const_iterator conIter = conn.begin();
      for (; conIter != conn.end(); conIter++) {
        pv = conIter->first;
        int locId = localIds[pv];
        int nbNeighbors = conIter->second.size();
        for (int iDir=0; iDir<3; iDir++) {
          int row = nbVert*iDir + locId;
          lsys->set_nnz(row,(nbNeighbors+1)*3);// A node is coupled with all its neighbors, and don't forget himself
        }
      }
    }

    lsys->allocate_matrix();
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::addVertexOnEdge(pVertex pv, pEdge pe)
  {
    if ( computed == true )
      {
        smallVector newDx(3);
        newDx.scale(0.);
        for (int i=0; i<2; i++)
          {
            pVertex vi = E_vertex(pe,i);
            std::map<pVertex,smallVector >::iterator iter = relocations.find(pv);
            if ( iter != relocations.end() ) {
              smallVector dxi = (*iter).second;
              newDx(0) += dxi(0); newDx(1) += dxi(1); newDx(2) += dxi(2);
            }
            else return;
          }
        newDx.scale(0.5);
        relocations[pv] = newDx;
      }
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::removeVertex(pVertex pv)
  {
    if ( computed == true ) 
      {
        std::map<pVertex,smallVector >::iterator iter = relocations.find(pv);
        if ( iter != relocations.end() ) {
          relocations.erase(iter);
        }
      }
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::setVertexResult(int nbVert, pVertex pv, 
                                        const MAdLinearSystemDef * lsys)
  {
    // get the vertex id
    map<pVertex,int>::const_iterator itId = localIds.find(pv);
    int id = (*itId).second;

    // get the dx
    smallVector dX(3);
    for (int iDir=0; iDir<3; iDir++) {
      int row = nbVert*iDir + id;
      dX(iDir) = lsys->getFromSolution(row);
    }

    relocations[pv] = dX;
  }
  
  // -------------------------------------------------------------------
  void MAdElasticityOp::setResults(int nbVert,
                                   const MAdLinearSystemDef * lsys)
  {
    relocations.clear();

    if (cavityEqualMesh) {
      pVertex pv;
      VIter vit = M_vertexIter(mesh);
      while ( ( pv = VIter_next(vit) ) ) {
        setVertexResult(nbVert,pv,lsys);
      }
      VIter_delete(vit);
    }
    else {
      set<pVertex> vertices;
      set<pRegion>::const_iterator cIter = cavity.begin();
      for (; cIter != cavity.end(); cIter++) {
        pPList rVerts = R_vertices(*cIter);
        void * temp = NULL;
        pVertex pv;
        while ( ( pv = (pVertex)PList_next(rVerts,&temp) ) ) {
          vertices.insert(pv);
        }
        PList_delete(rVerts);
      }

      set<pVertex>::const_iterator vIter = vertices.begin();
      for (; vIter != vertices.end(); vIter++) {
        setVertexResult(nbVert,*vIter,lsys);
      }
    }
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::printRelocations() const
  {
    std::cout << "Printing relocation: "<<relocations.size()<<" vertices relocated\n";
    std::map<pVertex,smallVector>::const_iterator iter = relocations.begin();
    for(; iter != relocations.end(); iter++)
      {
        pVertex pv = (*iter).first;
        smallVector dX = (*iter).second;
        cout <<"Vertex "<<pv<<" : "<<dX(0)<<" "<<dX(1)<<" "<<dX(2)<<"\n";
      }
    
    std::cout << "Printed relocation: "<<relocations.size()<<" vertices relocated\n";
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::printDirichlet() const
  {
    std::cout << "Printing dirichlet BCs: "<<dirichlet.size()<<" vertices\n";
    std::map<pVertex,smallVector>::const_iterator iter = dirichlet.begin();
    for(; iter != dirichlet.end(); iter++)
      {
        pVertex pv = (*iter).first;
        smallVector dX = (*iter).second;
        cout <<"Vertex "<<pv<<" : "<<dX(0)<<" "<<dX(1)<<" "<<dX(2)<<"\n";
      }
    
    std::cout << "Printed dirichlet BCs: "<<dirichlet.size()<<" vertices\n";
  }

  // -------------------------------------------------------------------
  int MAdElasticityOp::compute()
  {
    CPUTimeManager& tm = CPUTimeManagerSgl::instance();

    relocations.clear();

    // --------------------------------------------
    // If there is no element in the cavity, just 
    // prescribed node relocations (dirichlet) ...
    // --------------------------------------------
    if ( !cavityEqualMesh && cavity.empty() ) {
      MAdMsgSgl::instance().info(__LINE__,__FILE__,
                                 "Empty cavity for elastic computation");
      relocations = dirichlet;
    }

    // --------------------------------------------------
    // ... if an elastic computation has to be performed
    // --------------------------------------------------
    else {
      
      MeshQualityManagerSgl::instance().evaluateSizes(); // required for detJ0

      // --- produce the mapping pVertex -> localId ---
      int nbVert = generateVertexIds();

      // --- create the linear system ---
      double t_sys0 = tm.getTime();
      int sysSize = dim * nbVert;
      MAdLinearSystemDef * lsys = new MAdLinearSystemDef();
      lsys->allocate(sysSize);

      lsys->setSolver(CG);
      lsys->setFillIn(4);
      lsys->setEps(1.e-12);

      lsys->setPrec(1.e-12); // for gmm only
      lsys->setNoisy(1);    // for gmm only

      // --- allocate the matrix ---
      allocateMatrix(nbVert,lsys); // for PETSc only

      lsys->zeroMatrix();
      lsys->zeroRightHandSide();

      double dt_sys = tm.getTime() - t_sys0;
      MAdMsgSgl::instance().info(-1,__FILE__,
                                 "Created the system in %f seconds",dt_sys);

      // --- assemble the matrix ---
      double t_ass0 = tm.getTime();
      if (cavityEqualMesh) {
        RIter rit = M_regionIter(mesh);
        while ( pRegion pr = RIter_next(rit) ) addToMatrix(pr,nbVert,lsys);
        RIter_delete(rit);
      }
      else {
        set<pRegion>::const_iterator cavIter = cavity.begin();
        for (; cavIter != cavity.end(); cavIter++) addToMatrix(*cavIter,nbVert,lsys);
      }
      double dt_ass = tm.getTime() - t_ass0;
      MAdMsgSgl::instance().info(-1,__FILE__,
                                 "Assembled the matrix in %f seconds",dt_ass);

      // --- apply bc (ugly but it works hum) ---
      double t_bc0 = tm.getTime();
      map<pVertex,smallVector>::iterator dIter = dirichlet.begin();
      map<pVertex,smallVector>::iterator dLast = dirichlet.end();
      for (; dIter != dLast; dIter++) {

        pVertex pv       = (*dIter).first;
        smallVector dxyz = (*dIter).second;

        // get the vertex id
        map<pVertex,int>::const_iterator itId = localIds.find(pv);
        int id = (*itId).second;
    
        for (int iDir=0; iDir<3; iDir++) {
          int row = nbVert*iDir + id;
          const double BIG = /*lsys->getFromMatrix(row,row) *  */1.e12;
          lsys->addToMatrix(row,row,BIG);
          lsys->addToRightHandSide(row,BIG * dxyz(iDir));
        }
    
      }
      double dt_bc = tm.getTime() - t_bc0;
      MAdMsgSgl::instance().info(-1,__FILE__,
                                 "Applied boundary conditions in %f seconds",dt_bc);

      // --- reorder matrix ---
      lsys->reorder();

      // --- solve the system ---
      double t_slv0 = tm.getTime();
      int converged = lsys->systemSolve();
      double dt_slv = tm.getTime() - t_slv0;
      MAdMsgSgl::instance().info(-1,__FILE__,
                                 "Solved the elastic system (convergence flag: %d) in %f seconds",
                                 converged,dt_slv);


      // --- fill relocations with the results ---
      setResults(nbVert,lsys);

      localIds.clear();
      delete lsys;
    }

    dirichlet.clear();
    cavity.clear();

    computed = true;

    return 1;
  }


  // -------------------------------------------------------------------
  bool MAdElasticityOp::checkVolume(const pRegion pr, double ratio)
  {
    bool inCavity = false;

    double rxyz[4][3];
    R_coordP1(pr,rxyz);

    pPList rVerts = R_vertices(pr);
    void * temp = NULL;
    int i = 0;
    pVertex pv;
    while ( ( pv = (pVertex)PList_next(rVerts,&temp) ) ) {
      std::map<pVertex,smallVector>::const_iterator iter = relocations.find(pv);
      if ( iter != relocations.end() ) {
        inCavity = true;
        for (int j=0; j<3; j++) rxyz[i][j] += ratio * (*iter).second(j);
      }
      i++;
    }
    PList_delete(rVerts);

    // Do the check only if one of the nodes moved
    if ( inCavity && R_XYZ_volume(rxyz) <= MAdTOL ) return false;

    return true;
  }

  // -------------------------------------------------------------------
  // Cannot use 'cavity': it is hard to maintain if mesh modifications
  // occur during advancement so it is cleared in 'compute()'.
  bool MAdElasticityOp::checkVolumes(double ratio)
  {
//     if ( cavityEqualMesh ) {
    pRegion pr;
    RIter rit = M_regionIter(mesh);
    while ( ( pr = RIter_next(rit) ) ) {
      if ( !checkVolume(pr,ratio) ) {
// #warning "debug"
//         printf("Found a negative volume\n");
//         pPList lis = PList_new();
//         PList_append(lis,pr);
//         printPosEntities(lis,"neg_vol_in_elast.pos",OD_DIMENSION);
        RIter_delete(rit);
        return false;
      }
    }
    RIter_delete(rit);
//     }
//     else {
//       set<pRegion>::const_iterator cIter = cavity.begin();
//       for (; cIter != cavity.end(); cIter++) {
//         if ( !checkVolume(*cIter,ratio) ) return false;
//       }
//     }

    return true;
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::relocate(double ratio)
  {
    pVertex pv;
    double xyz[3];
    
    std::map<pVertex,smallVector>::const_iterator iter = relocations.begin();
    for(; iter != relocations.end(); iter++)
      {
        pv = (*iter).first;
        V_coord(pv,xyz);

        smallVector dX = (*iter).second;

        xyz[0] += ratio * dX(0);
        xyz[1] += ratio * dX(1);
        xyz[2] += ratio * dX(2);
        if ( !V_setPosition(pv,xyz) ) {
          MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                      "Could not relocate a node");
          
        }
      }
  }

  // -------------------------------------------------------------------
  // Advances the relocation as far as possible
  // Returns:
  //    - 0: no relocation possible
  //    - 1: advanced but not to the final position
  //    - 2: reached the full prescribed relocation
  int MAdElasticityOp::advance(double * ratio, double tolerance)
  {
    if ( *ratio >= 1.-tolerance ) {
      *ratio = 1.;
      return 2;
    }

    int achieved = 0; // the result returned

    int subIter = 0;
    const int nbMaxIter = 6;

    double margin = 1. - *ratio; // in [0;1] : the rest of the motion to be performed
    double curRatio = 1.;        // in [0;1] : the part of the margin that will be tested next
    double delta = 0.5;          // in [0;1] : the increment/decrement of ratio at next iteration
    double best = -1.;           // in [0;1] : the best current ratio with positive volumes

    while ( subIter < nbMaxIter )
      {
        double totalAdvancement = curRatio * margin + *ratio;
        if ( !checkVolumes( curRatio * margin ) ) {
          curRatio -= delta;
        }
        else {
          best = curRatio;
          if ( totalAdvancement + tolerance >= 1. ) { achieved = 2; break; }
          else                                      { achieved = 1; curRatio += delta; }
        }

        subIter++;
        delta *= 0.5;

        printf("Ratio: %f, margin: %f, achieved: %d, subIter: %d, best: %f, curRatio: %f, delta: %f\n",
               *ratio, margin, achieved, subIter, best, curRatio, delta);
      }

    double moveRatio = 0.;
    if ( achieved == 2 ) {
      moveRatio = ( 1. - *ratio );
      relocate(moveRatio);
      *ratio = 1.;
    }
    else if ( achieved == 1 ) {
      moveRatio = best * margin;
      relocate(moveRatio);
      *ratio += moveRatio;
    }
    
    return achieved;
  }

  // -------------------------------------------------------------------
  // Force the relocation even if it leads to negative volumes
  void MAdElasticityOp::forceRelocation()
  {
    relocate(1.);
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::element3DMatrix(pRegion pr, smallMatrix& m) const
  {
    int nbNodes = 4;
    double xyz[4][3];
    R_coordP1(pr,xyz);

    double FACT = E / (1 + nu);
    double C11 = FACT * (1 - nu) / (1 - 2 * nu);
    double C12 = FACT * nu / (1 - 2 * nu);
    double C44 = (C11 - C12) / 2;
    const double C[6][6] =
      { {C11, C12, C12,    0,   0,   0}, 
        {C12, C11, C12,    0,   0,   0}, 
        {C12, C12, C11,    0,   0,   0}, 
        {  0,   0,   0,  C44,   0,   0},
        {  0,   0,   0,    0, C44,   0}, 
        {  0,   0,   0,    0,   0, C44} };

    smallMatrix H   (6,6);
    for (int i=0;i<6;i++)
      for (int j=0;j<6;j++)
        H(i,j) = C[i][j];

    smallMatrix B   (6,3*nbNodes);
    smallMatrix BTH (3*nbNodes,6);
    smallMatrix BT  (3*nbNodes,6);

    const double gradShpFct[4][3] =
      { {-1., -1., -1.},
        { 1.,  0.,  0.},
        { 0.,  1.,  0.},
        { 0.,  0.,  1.} };

    double jac [3][3];
    jac[0][0] = jac[0][1] = jac[0][2] = 0.;
    jac[1][0] = jac[1][1] = jac[1][2] = 0.;
    jac[2][0] = jac[2][1] = jac[2][2] = 0.;
    for(int i = 0; i < nbNodes; i++) {
      jac[0][0] += xyz[i][0] * gradShpFct[i][0]; jac[0][1] += xyz[i][1] * gradShpFct[i][0]; jac[0][2] += xyz[i][2] * gradShpFct[i][0];
      jac[1][0] += xyz[i][0] * gradShpFct[i][1]; jac[1][1] += xyz[i][1] * gradShpFct[i][1]; jac[1][2] += xyz[i][2] * gradShpFct[i][1];
      jac[2][0] += xyz[i][0] * gradShpFct[i][2]; jac[2][1] += xyz[i][1] * gradShpFct[i][2]; jac[2][2] += xyz[i][2] * gradShpFct[i][2];
    }
    const double detJ = fabs(jac[0][0] * jac[1][1] * jac[2][2] + jac[0][2] * jac[1][0] * jac[2][1] +
                             jac[0][1] * jac[1][2] * jac[2][0] - jac[0][2] * jac[1][1] * jac[2][0] -
                             jac[0][0] * jac[1][2] * jac[2][1] - jac[0][1] * jac[1][0] * jac[2][2]);

    double invjac [3][3];
    inverseMat (jac, invjac) ;

    double Grads[4][3];

    B.set_all(0.0);
    BT.set_all(0.0);

    for (int j=0;j<nbNodes;j++){

      Grads[j][0] = invjac[0][0] * gradShpFct[j][0] + invjac[0][1] * gradShpFct[j][1] + invjac[0][2] * gradShpFct[j][2];
      Grads[j][1] = invjac[1][0] * gradShpFct[j][0] + invjac[1][1] * gradShpFct[j][1] + invjac[1][2] * gradShpFct[j][2];
      Grads[j][2] = invjac[2][0] * gradShpFct[j][0] + invjac[2][1] * gradShpFct[j][1] + invjac[2][2] * gradShpFct[j][2];
    
      BT(j,0) = B(0,j) = Grads[j][0];
      BT(j,3) = B(3,j) = Grads[j][1];
      BT(j,4) = B(4,j) = Grads[j][2];
    
      BT(j+nbNodes,1) = B(1,j+nbNodes) = Grads[j][1];
      BT(j+nbNodes,3) = B(3,j+nbNodes) = Grads[j][0];
      BT(j+nbNodes,5) = B(5,j+nbNodes) = Grads[j][2];
    
      BT(j+2*nbNodes,2) = B(2,j+2*nbNodes) = Grads[j][2];
      BT(j+2*nbNodes,4) = B(4,j+2*nbNodes) = Grads[j][0];
      BT(j+2*nbNodes,5) = B(5,j+2*nbNodes) = Grads[j][1];
    }
  
    BTH.set_all(0.0);
    BTH.blas_dgemm (BT,H); 
  
    m.set_all(0.0);
    m.blas_dgemm (BTH,B,detJ,1.0);

    // --- Selective element stiffening ---
    // reference:
    // Stein, Tezduyar, Benney "Mesh Moving Techniques for FSI with Large Displacement", ASME, 2003.

    double detJ0 = MeshQualityManagerSgl::instance().getMaxSize();
    if ( detJ0 <= 0. ) detJ0 = 1.;

    double scaleFactor = pow( detJ0 / detJ, chi);
    if ( scaleFactor > 1.e12 || scaleFactor < 1.e-12 ) {
      cout <<"Warning in MAdElasticity: scale factor for element stiffness is "<<scaleFactor<<endl;
    }

    m.scale ( scaleFactor );
  }

  // -------------------------------------------------------------------

}
