// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

// the template parameter file for this executable
#include "Parameters.h"
// the parameter file for the current test case
#ifdef _HAVE_PARSER_
 #include "moveItParse.h"
#else
 #include "MyParams.h"
#endif

#include "MAdLib.h"
#include "NullSField.h"

#include <iostream>
#include <sstream>
#include <math.h>
#include <sys/time.h>
#include <fstream>
#include <stdlib.h>
#include <vector>
#include <set>

#ifdef PARALLEL
 #include "mpi.h"
#endif

using std::cout;
using std::cerr;
using std::endl;
using std::stringstream;
using std::ofstream;
using std::vector;
using std::set;

using namespace MAd;

// ----------------------------------------------------------------------
double CPUTime() {
  
#ifdef PARALLEL
  return MPI_Wtime();
#else
  struct timeval tp;
  struct timezone tz;

  gettimeofday(&tp,&tz);

  return ((double) tp.tv_sec +
          (double) ((double) .000001 * (double) tp.tv_usec));
#endif
}

// ----------------------------------------------------------------------
void writeSolution(MeshAdapter* ma, int iter, string type) 
{
  stringstream ss;
  string iterStr;  ss << iter;  ss >> iterStr;
  
  if ( !strcmp(type.c_str(),"Pos") || !strcmp(type.c_str(),"MshAndPos") ) {
    string namePos = "result" + iterStr + ".pos";
    ma->writePos(namePos,OD_ISOSIZEFIELD);
  }
  
  if ( !strcmp(type.c_str(),"Msh") || !strcmp(type.c_str(),"MshAndPos") ) {
    string nameMsh = "result" + iterStr + ".msh";
    ma->writeMsh(nameMsh);
  }
}

// ----------------------------------------------------------------------
void setMobileObject(mobileObject* mob, pMesh mesh, ObjectDef def)
{
  // --- Name ---
  string name = def.name;
  mob->setName(name);

  // --- Geometry ---
  int type = def.geomLevel;
  set<int>::const_iterator gIter = def.geomTags.begin();
  set<int>::const_iterator gLast = def.geomTags.end();
  for (; gIter != gLast; gIter++) {
    for (int iTag = (*gIter); iTag <= (*gIter); iTag++) {
      mob->addGEntity(type,iTag);
    }
  }

  // --- Kinematics ---
  KinematicType kinType = def.kinType;
  if ( kinType == KT_DISPLACEMENT ) {
    mob->setDxKinematics(def.kinExpression);
  }
  else {
    mob->setVKinematics(PARSED, NULL, NULL, def.kinExpression);
  }

  // --- Local size fields ---
  std::list<LocalSFDef>::const_iterator sIter = def.sizes.begin();
  std::list<LocalSFDef>::const_iterator sLast = def.sizes.end();
  for (; sIter != sLast; sIter++) {
    double radius = (*sIter).radius;
    double sizeIn = (*sIter).sizeIn;
    double sizeOut = (*sIter).sizeOut;
    double scale = sizeOut / sizeIn;
    LocalSizeField* locSField = new LocalSizeField(mesh,radius,sizeIn,scale);
    set<int>::const_iterator gIter = def.geomTags.begin();
    set<int>::const_iterator gLast = def.geomTags.end();
    for (; gIter != gLast; gIter++) {
      for (int iTag = (*gIter); iTag <= (*gIter); iTag++) {
        locSField->addGeometricEntity(type,iTag);
      }
    }
    locSField->updateTree();
    mob->addLocalSField(locSField);
  }
}

// ----------------------------------------------------------------------
void deleteObjects(mobileObjectSet* objSet)
{
  set<mobileObject*> objs = objSet->getObjects();
  set<mobileObject*>::iterator oIter = objs.begin();
  for (; oIter != objs.end(); oIter++) {
    
    mobileObject * mob = *oIter;

    // clean size fields
    set<LocalSizeField* > localSFs = mob->getSizes();
    set<LocalSizeField* >::iterator sIter = localSFs.begin();
    for (; sIter != localSFs.end(); sIter++) {
      if (*sIter) delete (*sIter);
    }
    
    // delete object
    if (mob) delete mob;
  }
}

// ----------------------------------------------------------------------
int main(int argc, char* argv[]) 
{
  MAdLibInitialize(argc,argv);

#ifdef _HAVE_PARSER_
  parseControlFile(argc,argv);
#endif

  MoveParameters parameters = getParameters();

  // ------------------------------------------------
  // setup the output
  // ------------------------------------------------

  string outputType = parameters.outType;
  int outputFrequency = parameters.outFrequency;
  string outputPrefix = parameters.outPrefix;

  // ------------------------------------------------
  // load the mesh
  // ------------------------------------------------

  cout << "Loading the mesh...\n";
  double cpu_mesh_0 = CPUTime();

  // --- Reading model ---
  string meshFileName = parameters.meshFileName;
  string geoFileName  = parameters.geoFileName;
  pGModel model = 0;
  GM_create(&model,"theModel");
  if ( !geoFileName.empty() ) GM_read(model, geoFileName.c_str());
  else                        GM_readFromMSH(model, meshFileName.c_str());

  // --- Reading mesh ---
  pMesh mesh = M_new(model);
  M_load(mesh,meshFileName.c_str());
//   M_writeMsh (mesh, (outputPrefix + "initMesh.msh").c_str(), 1, NULL);

  double cpu_mesh_tot = CPUTime() - cpu_mesh_0;
  cout << "Loaded the mesh in "<<cpu_mesh_tot<<" seconds\n";

  // ------------------------------------------------
  // build the mesh adapter
  // ------------------------------------------------

  cout << "Building the mesh adapter...\n";
  double cpu_ma_0 = CPUTime();
  MeshAdapter* ma = new MeshAdapter(mesh);

  ma->setOutputPrefix(outputPrefix);

  ma->setAlgorithm(CPS_SWP_SLV_SPT);
  ma->setMaxIterationsNumber(parameters.maxInnerIter);

  double lowerLenghtBound = parameters.lowerLength;
  double upperLenghtBound = parameters.upperLength;
  ma->setEdgeLenSqBounds( lowerLenghtBound * lowerLenghtBound,
                          upperLenghtBound * upperLenghtBound );

  ma->setSwapMinImproveRatio (parameters.swapMinImproveRatio);
  ma->setSliverQuality       (parameters.sliverQuality);
  ma->setSliverPermissionInESplit (parameters.splitCanMakeSlivers);
  ma->setSliverPermissionInECollapse (parameters.collapseCanMakeSlivers);
  ma->setCollapseOnBoundary  (parameters.collapseOnBoundary,
                              parameters.clpOnBdryTolerance);
  ma->setSwapOnBoundary      (parameters.swapOnBoundary,
                              parameters.swapOnBdryTolerance);
  ma->setGeoTracking         (parameters.trackGeometry,
                              parameters.snap_cavityIsMesh,
                              parameters.snap_thickness,
                              parameters.snap_chi);
  ma->setSizeFieldSmoothing  (parameters.SFSmoothing,
                              parameters.SFSmoothGrad);

  double cpu_ma_tot = CPUTime() - cpu_ma_0;
  cout << "Built the mesh adapter in "<<cpu_ma_tot<<" seconds\n";

  // ------------------------------------------------
  // build the size fields
  // ------------------------------------------------
  cout << "Buidling the size fields...\n";
  double cpu_sf_0 = CPUTime();

  set<pSField> sizeFields;

  std::list<AnalSFDef>::const_iterator sIter = parameters.sizes.begin();
  std::list<AnalSFDef>::const_iterator sLast = parameters.sizes.end();
  for (; sIter != sLast; sIter++) {

    SizeFieldBase* sizeField = NULL;
  
    OrientationType orient = (*sIter).orientation;
    if ( orient == ORT_ISOTROPIC ) {
      sizeField = new AnalyticalSField((*sIter).isoSize);
    }
    else {
      sizeField = new AnalyticalSField((*sIter).anisoSize,
                                       (*sIter).anisoDir0,
                                       (*sIter).anisoDir1,
                                       (*sIter).anisoDir2);
    }

    sizeFields.insert(sizeField);
    ma->addSizeField(sizeField);
  }

  double cpu_sf_tot = CPUTime() - cpu_sf_0;
  cout << "Built the size field in "<<cpu_sf_tot<<" seconds\n";

  ma->printStatistics(cout);

  // #warning "debug"
  //   ma->writePos("testSFInit.pos",OD_ISOSIZEFIELD);

  // ------------------------------------------------
  // setup the quality monitoring
  // ------------------------------------------------

  string qualityFileName = outputPrefix + "quality";
  FILE* qualFile = fopen(qualityFileName.c_str(),"w");
  if (!qualFile) {
    cerr << "Could not open the quality file " << qualityFileName << endl;
    throw;
  }

  double meanShape, worstShape;
  ma->getStatistics(&meanShape,&worstShape);
  fprintf(qualFile,"iter\tmean quality\tworst quality\n");
  fprintf(qualFile,"%d\t%g\t%g\n",0,meanShape,worstShape);
  fclose(qualFile);

  // ------------------------------------------------
  // setup the cpu time monitoring
  // ------------------------------------------------

  string cpuFileName = outputPrefix + "cpu";
  FILE* cpuFile = fopen(cpuFileName.c_str(),"w");
  if (!cpuFile) {
    cerr << "Could not open the cpu report file " << cpuFileName << endl;
    throw;
  }

  fprintf(cpuFile,"iter\tnodes motion\tadaptation\toutputs  \taccumulated nm\taccum adapt\taccum out\ttotal\n");
  fclose(cpuFile);

  string produceCpuInfo = "cat /proc/cpuinfo > \"" + outputPrefix + "cpuinfo\"";
  system(produceCpuInfo.c_str());

  // ------------------------------------------------
  // setup the mesh size monitoring
  // ------------------------------------------------

  string MSFileName = outputPrefix + "meshSize";
  FILE* MSFile = fopen(MSFileName.c_str(),"w");
  if (!MSFile) {
    cerr << "Could not open the mesh size report file " << MSFileName << endl;
    throw;
  }

  fprintf(MSFile,"iter\t#nodes\t#edges\t#faces\t#regions\n");
  fprintf(MSFile,"init\t%d\t%d\t%d\t%d\n",
          M_numVertices(mesh),M_numEdges(mesh),M_numFaces(mesh),M_numRegions(mesh));
  fclose(MSFile);

  // ------------------------------------------------
  // setup debug tools
  // ------------------------------------------------

  int debugLevel = parameters.debugLevel;
  ma->setDebugLevel(debugLevel);

  // --- cross checks between two executions ---
  bool openJournal = parameters.openJournal;
  if ( openJournal ) {
    
    // Will hold a list of operations and checks in this execution
    ma->openJournal();

    // Will compare the list to another one (and abort if different)
    string refJournalName = parameters.referenceJournal;
    if ( !refJournalName.empty() ) {
      std::cout<<"Warning: comparing execution with the journal \'"
               << refJournalName << "\'\n";
      ma->setReferenceJournal(refJournalName);
    }
  }

  // --- slivers output ---
  bool sliverReports = parameters.sliverReports;
  if ( sliverReports ) {
    ma->enableSliverReports();
  }
  bool sliverOps = parameters.testSliverOperators;
  ma->testSliverOperators(sliverOps);

  // ------------------------------------------------
  // TEST MOBILE OBJECTS
  // ------------------------------------------------
  
  if ( !strcmp(parameters.task.c_str(),"MobileObject") ) {

    ma->storeInitialCoordinates();

    // ---- build the mobile objects ----
    cout << "Building the mobile objects...\n";
    double cpu_mobj_0 = CPUTime();
    mobileObjectSet* objs = new mobileObjectSet();
    for (unsigned int iObj = 0; iObj < parameters.objects.size(); iObj++) {
      mobileObject* mob = new mobileObject(mesh);
      setMobileObject(mob,mesh,parameters.objects[iObj]);
      objs->insert(mob);
    }
    ma->registerObjects(objs);
    double cpu_mobj_tot = CPUTime() - cpu_mobj_0;
    cout << "Built the mobile objects in "<<cpu_mobj_tot<<" seconds\n";

    // #warning "Size field printed to debug"
    //   ma->writePos("sizeFieldInit.pos",OD_ISOSIZEFIELD);

    // ---- get node motion parameters ----
    string nodeMotionType = parameters.nodeMotionType;
    double elasticChi = parameters.elasticStiffnessAlteration;
    bool   elasticCavityMesh = parameters.elasticIsMeshTheCavity;
    int    elasticCavityThick = parameters.elasticCavityThickness;

    // ---- get time parameters ----
    double dtGlob = parameters.timeStep;
    double finalTime = parameters.finalTime;
    int maxTimeSteps = parameters.maxNbTimeSteps;

    // ---- the loop ----
    double cpu_t0 = CPUTime();
    double t = 0., dt_iter = 0.;
    int iter = 0;
    double cpu_move = 0., cpu_adapt = 0., cpu_output = 0.;
    while (iter <= maxTimeSteps && t < finalTime)
      {
        double cpu_t1 = CPUTime();
        if ( iter > 0 )  // a first step is performed with the adaptation before any motion occurs
          {
            // --- find time step for this iteration ---
            double t_left_tot = std::max(finalTime - t, 0.0);
            dt_iter = std::min(dtGlob, t_left_tot);
          
            // --- move nodes (dummy move and Laplace smoothing) ---
            if ( !strcmp(nodeMotionType.c_str(),"ForceBoundaries") ) {
              double part;
              if (!ma->partlyMoveObjects(t, dt_iter, &part)) {
                cout <<"Could not perform a forced move\n";
                ma->abort(__LINE__,__FILE__);
              }
              dt_iter *= part;

              if ( debugLevel >= 1 && !(ma->checkTheMesh()) )  ma->abort(__LINE__,__FILE__);
              ma->LaplaceSmoothing(OPTIMAL);
              if ( debugLevel >= 1 && !(ma->checkTheMesh()) )  ma->abort(__LINE__,__FILE__);
            }

            t += dt_iter;  ma->setTime(t);
            printf("\nGlobal time step %d, t = %f, dt = %f\n\n",iter,t,dt_iter);

            // --- move nodes (elasticity analogy) ---
            if ( !strcmp(nodeMotionType.c_str(),"ElasticAnalogy") ) {
              ma->moveObjectsAndReposition(t, dt_iter, 
                                           elasticChi,
                                           elasticCavityMesh, 
                                           elasticCavityThick);
            }
          }
        double cpu_t2 = CPUTime(); 
        cpu_move += ( cpu_t2 - cpu_t1 );

        // --- adapt the mesh ---
        ma->run();
        double cpu_t3 = CPUTime();
        cpu_adapt += ( cpu_t3 - cpu_t2 );

        // --- outputs ---

        {  // full mesh
          if ( (iter%outputFrequency) == 0 ) {
            writeSolution(ma,iter,outputType);
          }
        }

        {  // quality monitor
          ma->getStatistics(&meanShape,&worstShape);
          FILE* qf = fopen(qualityFileName.c_str(),"a");
          fprintf(qf,"%d\t%e\t%e\n",iter,meanShape,worstShape);
          fclose(qf);
        }

        {  // cpu monitor
          double cpu_t4 = CPUTime();
          cpu_output += ( cpu_t4 - cpu_t3 );
      
          FILE* cpuF = fopen(cpuFileName.c_str(),"a");
          fprintf(cpuF,"%d\t%e\t%e\t%e\t%e\t%e\t%e\t%e\n",
                  iter, cpu_t2 - cpu_t1, cpu_t3 - cpu_t2, cpu_t4 - cpu_t3,
                  cpu_move, cpu_adapt, cpu_output, cpu_t4 - cpu_t0 );
          fclose(cpuF);
        }

        {  // mesh size monitor
          FILE* msf = fopen(MSFileName.c_str(),"a");
          fprintf(msf,"%d\t%d\t%d\t%d\t%d\n",
                  iter,M_numVertices(mesh),M_numEdges(mesh),M_numFaces(mesh),M_numRegions(mesh));
          fclose(msf);
        }

        { // slivers stats
          if ( (iter%10) == 0 ) {
            ofstream sliverOut( (outputPrefix + "slivers_tmp").c_str() );
            ma->printSliverRegionStatistics(sliverOut);
          }
        }

        { // operations stats
          if ( (iter%10) == 0 ) {
            ofstream operationsOut( (outputPrefix + "statistics_tmp").c_str() );
            ma->printStatistics(operationsOut);
          }
        }

        { // journal
          if ( (iter%10) == 0 ) {
            ofstream journalOut( (outputPrefix + "journal_tmp").c_str() );
            ma->flushJournal(journalOut);
          }
        }

        iter++;
      }

    if (objs) {
      deleteObjects(objs);
      delete objs;
    }

    ma->removeStoredCoordinates();
  }

  // ------------------------------------------------
  // TEST REMOVE REGION
  // ------------------------------------------------

  else if ( !strcmp(parameters.task.c_str(),"RemoveRegion") )
    {
      for (int i=0; i<1; i++) {
        int numOp = 0;
        int oriNumFaces = M_numFaces(mesh);
        int count = 0;
        FIter fi = M_faceIter(mesh);
        pFace pf;
        while ( ( pf = FIter_next(fi) ) && ( count < oriNumFaces ) ) {
          count++;
          if (F_numRegions(pf) != 1) continue;
          else {
            pRegion pr = F_region(pf,0);
            if (!pr) pr = F_region(pf,1);
            if ( ma->removeRegion(pr) ) numOp++;
          }
        }
        FIter_delete(fi);
        cout << "Num region removed: "<<numOp<<endl;
      }
    }

  // ------------------------------------------------
  // test size fields intersection or smoothing
  // ------------------------------------------------
  
  else if ( !strcmp(parameters.task.c_str(),"SizeField") )
    {
      PWLSField* testSF = new PWLSField(mesh);
      string h = "2.*x+0.1";
      AnalyticalSField* testASF = new AnalyticalSField(h);

      testSF->intersect(testASF);
  
      MeshAdapter* testMA = new MeshAdapter(mesh,testSF);
  
      testMA->writePos("sizeField.pos",OD_ISOSIZEFIELD);
      testMA->writePos("meanRatio.pos",OD_MEANRATIO);

      testSF->smooth(1.);
      testMA->writePos("sizeFieldSmoothed.pos",OD_ISOSIZEFIELD);

      exit(0);
    }

  // ------------------------------------------------
  // TEST REMOVE SLIVER FACES
  // ------------------------------------------------

  else if ( !strcmp(parameters.task.c_str(),"RemoveSliverFaces") )
    {
      for (int i=0; i<10; i++) {
        ma->removeSlivers();
        if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__);
      }
    }

  // ------------------------------------------------
  // TEST LAPLACE SMOOTHING
  // ------------------------------------------------
  
  else if ( !strcmp(parameters.task.c_str(),"LaplaceSmoothing") )
    {
      for (int i=0; i<1; i++) {
        ma->LaplaceSmoothing(FAST);
        cout << "Smoothing "<<i<<" operated\n";
        if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__);
      }
    }

  // ------------------------------------------------
  // TEST GEOMETRIC CONSTRAINTS
  // ------------------------------------------------
  
  else if ( !strcmp(parameters.task.c_str(),"GeoConstraints") )
    {
      ma->setConstraint(2,5);
      ma->removeConstraint(2,5);
    }

  // ------------------------------------------------
  // TEST CONSTRAINTS
  // ------------------------------------------------

  else if ( !strcmp(parameters.task.c_str(),"TopoConstraints") )
    {
      FIter fi = M_faceIter(mesh);
      while ( pFace pf = FIter_next(fi) ) {
        int gtag  = GEN_tag ( EN_whatIn ((pEntity)pf) );
        int gtype = GEN_type( EN_whatIn ((pEntity)pf) );
        if ( gtype == 2 && ( gtag == 5 || gtag == 22 || gtag == 14 || gtag == 27 ) ) ma->setConstraint((pEntity)pf);
      } 
      FIter_delete(fi);

      VIter vi = M_vertexIter(mesh);
      pVertex pv;
      while ( ( pv = VIter_next(vi) ) ) {
        // int gtag  = GEN_tag ( EN_whatIn ((pEntity)pv) );
        int gtype = GEN_type( EN_whatIn ((pEntity)pv) );
        if ( gtype == 0 ) EN_constrain((pEntity)pv);
      } 
      VIter_delete(vi);
    }

  // ------------------------------------------------
  // TEST FACE SWAP OPERATOR
  // ------------------------------------------------

  else if ( !strcmp(parameters.task.c_str(),"FaceSwap") )
    {
      for (int i=0; i<5; i++) {
        int count=0, numOp=0;
        int oriNumFaces = M_numFaces(mesh);
        FIter fi = M_faceIter(mesh);
        pFace pf;
        while ( ( pf = FIter_next(fi) ) && ( count < oriNumFaces ) ) {
          int res = ma->swapFace(pf);
          if ( res ) numOp++;
          count++;
        }
        FIter_delete(fi);
        cout << "Num face swaps: "<<numOp<<endl;
        if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__);

        int nESwaps = ma->optimiseElementShape();
        cout<< "Edge swaps applied: "<<nESwaps<<endl;
      }
    }

  // ------------------------------------------------
  // MAKE A BAD MESH
  // ------------------------------------------------

  else if ( !strcmp(parameters.task.c_str(),"UglyMesh") )
    {
      ma->uglyTheMesh(0.2,10);
      ma->writeMsh("uglyMesh.msh");
    }

  // ------------------------------------------------
  // TEST REMOVE SLIVER REGIONS
  // ------------------------------------------------

  else if ( !strcmp(parameters.task.c_str(),"RemoveSliverRegions") )
    {
      ma->setSliverQuality(0.1);
      for (int i=0; i<2; i++) {
        ma->removeSlivers();
        if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__);
      }
    }
  
  // ------------------------------------------------
  // TEST DOUBLE-SPLIT-COLLAPSE OPERATOR
  // ------------------------------------------------

  else if ( !strcmp(parameters.task.c_str(),"DESC") )
    {
      for (int i=0; i<2; i++) {
        int numOp = 0;
        int oriNumRgn = M_numRegions(mesh);
        int count = 0;
        RIter ri = M_regionIter(mesh);
        pRegion pr;
        while ( ( pr = RIter_next(ri) ) && ( count < oriNumRgn ) ) {
          pEdge pe1 = R_edge(pr,0);
          pEdge pe2 = R_gtOppEdg(pr, pe1);

          if ( ma->DSplitCollapseEdge(pr,pe1,pe2) )
            numOp++;
          count++;
        }
        RIter_delete(ri);
        cout << "Num double-split-collapses: "<<numOp<<endl;

        if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__);
        //     ma->printStatistics(std::cout);

        //     int* nbBefore = new int(0);
        //     int* nbAfter = new int(0);
        //     ma->removeSliverRegions(nbBefore,nbAfter);
        //     cout <<"Nb slivers before->after realignement:\t"<<*nbBefore<<" -> "<<*nbAfter<<endl;
      }
    }

  // ------------------------------------------------
  // TEST EDGE SWAP
  // ------------------------------------------------
  
  else if ( !strcmp(parameters.task.c_str(),"EdgeSwap") )
    {
      if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__); 
      for (int i=0; i<10; i++) {
        int count = 0;
        int numOp = 0;
        int numE = M_numEdges(mesh);
        EIter ei = M_edgeIter(mesh);
        while ( pEdge pe = EIter_next(ei) ) {
          count++;
          if ( ma->swapEdge(pe) ) numOp++;
          if ( count > numE ) break;
        } 
        EIter_delete(ei);
        cout << "Num edge swaps ("<<i<<"): " << numOp << endl;
        if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__); 
        writeSolution(ma,i,outputType);
      }
    }

  // ------------------------------------------------
  // TEST GEOMATCHER
  // ------------------------------------------------
  
  else if ( !strcmp(parameters.task.c_str(),"GeoMatcher") )
    {
      for (int i=0; i<8; i++) {
        int count = 0;
        int numOp = 0;
        int numE = M_numEdges(mesh);
        EIter ei = M_edgeIter(mesh);
        pEdge pe;
        while ( ( pe = EIter_next(ei) ) ) {
          count++;
          if ( ma->splitEdge(pe) ) {
            ma->snapVertices();
            numOp++;
          }
          if ( count > numE ) break;
        } 
        EIter_delete(ei);
        cout << "Num edge splits ("<<i<<"): " << numOp << endl;
        if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__); 
      }
    }

  // ------------------------------------------------
  // TEST EDGE SPLIT
  // ------------------------------------------------
  
  else if ( !strcmp(parameters.task.c_str(),"EdgeSplit") )
    {
      if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__); 
      int numOp;
      for (int i=0; i<3; i++) {
        int count = 0;
        numOp = 0;
        int numE = M_numEdges(mesh);
        EIter ei = M_edgeIter(mesh);
        pEdge pe;
        while ( ( pe = EIter_next(ei) ) ) {
          count++;
//           if ( ma->splitEdge(pe) ) { numOp++; }
          if ( E_whatInType(pe) <= 2 && ma->splitEdge(pe) ) { numOp++; }
          if ( count > numE ) break;

// #warning "debug"
//           if (numOp>=7) break; 
        } 
        EIter_delete(ei);
        cout << "Num edge splits ("<<i<<"): " << numOp << endl;
        if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__); 
        writeSolution(ma,i,outputType);
      }
      // --- Geometry tracking ---
      if (numOp) ma->snapVertices();
    }

  // ------------------------------------------------
  // TEST EDGE COLLAPSE
  // ------------------------------------------------
  
  else if ( !strcmp(parameters.task.c_str(),"EdgeCollapse") )
    {
      GM_writeGEO(model,"testgeo.geo");
      ma->writeMsh("MeshInit.msh");
      for (int i=0; i<10; i++) {
        double t0 = CPUTime();
        int count = 0;
        int nbEdgesInit = M_numEdges(mesh);
        int numOp = 0;
        EIter ei = M_edgeIter(mesh);
        pEdge pe;
        while ( ( pe = EIter_next(ei) ) ) {
          count++;
          if ( ma->collapseEdge(pe) ) { numOp++; }
          if ( count > nbEdgesInit ) break;
        } 
        EIter_delete(ei);
        double dt = CPUTime() - t0;
        cout << "Num edge collapses ("<<i<<"): " << numOp << " in "<<dt<<" seconds\n";
        if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__); 
//         ma->printStatistics(std::cout);
        writeSolution(ma,i,outputType);
      }
    }

  // ------------------------------------------------
  // TEST EDGE SPLIT AND COLLAPSE
  // ------------------------------------------------
  
  else if ( !strcmp(parameters.task.c_str(),"EdgeSplitCollapse") )
    {
      for (int i=0; i<2; i++) {
        int count = 0;
        int oriNumEdges = M_numEdges(mesh);
        int numClp = 0, numSpl = 0;
        EIter ei = M_edgeIter(mesh);
        pEdge pe;
        while ( ( pe = EIter_next(ei) ) ) {
          count++;
          if ( (count%2) == 0 ) {
            if ( ma->collapseEdge(pe) ) numClp++;
          }
          else {
            if ( ma->splitEdge(pe) ) numSpl++;
          }
          if ( count > oriNumEdges ) break;
        } 
        EIter_delete(ei);
        cout << "Num edge splits ("<<i<<"): " << numSpl << endl;
        cout << "Num edge collapses ("<<i<<"): " << numClp << endl;
        if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__); 
        ma->printStatistics(std::cout);

        { // journal
          ofstream journalOut( (outputPrefix + "journal_tmp").c_str() );
          ma->flushJournal(journalOut);
        }
      }
    }

  // ------------------------------------------------
  // TEST FACE COLLAPSE OPERATOR
  // ------------------------------------------------

  else if ( !strcmp(parameters.task.c_str(),"FaceCollapse") )
    {
      for (int i=0; i<2; i++) {
        int numOp = 0;
        int oriNumFaces = M_numFaces(mesh);
        int count = 0;
        FIter fi = M_faceIter(mesh);
        pFace pf;
        while ( ( pf = FIter_next(fi) ) && ( count < oriNumFaces ) ) {
          for (int j=0; j<3; j++) {
            pEdge pe = F_edge(pf,j);
            if ( ma->collapseFace(pf,pe) ) {
              numOp++;
              break;
            }
          }
          count++;
        }
        FIter_delete(fi);
        cout << "Num split-collapses: "<<numOp<<endl;
        if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort(__LINE__,__FILE__);
        ma->printStatistics(std::cout);
      }
    }

  // ------------------------------------------------
  // TEST OPTIMISE EDGE LENGTH
  // ------------------------------------------------

  else if ( !strcmp(parameters.task.c_str(),"OptimizeLength") )
    {
      cout<<"Split collapses applied: "<<ma->optimiseEdgeLength()<<endl;
    }

  // ------------------------------------------------
  // TEST OPTIMISE SHAPE
  // ------------------------------------------------

  else if ( !strcmp(parameters.task.c_str(),"OptimizeShape") )
    {
      int nSwaps = ma->optimiseElementShape();
      cout<< "Swaps applied: "<<nSwaps<<endl;
    }

  // ------------------------------------------------
  else if ( !strcmp(parameters.task.c_str(),"None") )
    {
    }

  else {
    MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                "Unknown task %s",parameters.task.c_str());
  }

  // ------------------------------------------------
  // Write final result
  // ------------------------------------------------

  ma->writeMsh("result.msh");
  ma->writePos("result.pos",OD_MEANRATIO);

  ma->printStatistics(cout);
  ofstream statOut((outputPrefix + "statistics").c_str());
  ma->printStatistics(statOut);

  ofstream sliverOut((outputPrefix + "slivers").c_str());
  ma->printSliverRegionStatistics(sliverOut);

  ofstream journalOut((outputPrefix + "journal").c_str());
  ma->flushJournal(journalOut);

  // ------------------------------------------------
  // Clean up
  // ------------------------------------------------

  if (ma) delete ma;
  set<pSField>::iterator sfIter = sizeFields.begin();
  set<pSField>::iterator sfLast = sizeFields.end();
  for (; sfIter != sfLast; sfIter++) {
    delete (*sfIter);
  }
  
  if (mesh)  M_delete(mesh);
  if (model) GM_delete(model);

  MAdLibFinalize();
}

// ----------------------------------------------------------------------
