// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Jean-Francois Remacle, Gaetan Compere
// -------------------------------------------------------------------

#include "CheckMesh.h"
#include "MAdMessage.h"

#include <stdio.h>

namespace MAd {

  // -------------------------------------------------------------------
  int checkRegionsVolume(MDB_Mesh * mesh, int verbose, std::ostream& out) {

    int flag = 1;
    pRegion r;
    RIter rIt= M_regionIter(mesh);
    while ( ( r = RIter_next(rIt) ) ) {
      if (R_volume(r) < 0.) {
        flag = 0;
        if (verbose) {
          out << "Negative volume found !\n";
          R_info_topology(r,out);
        }
        break;
      }  
    }
    RIter_delete(rIt);
    return flag;
  }

  // -------------------------------------------------------------------
  int checkEdgeToRegionConnectivity(MDB_Mesh * mesh, int verbose, 
                                    std::ostream& out)
  {
#ifdef PARALLEL
    // In a mesh partitioned with MDB, the parallel boundary 
    // faces are not classified on a model face and the following 
    // check fails
    return 1;
#endif

    if ( M_dim(mesh) <= 2 ) return 1;

    int flag = 1;
    pEdge edge;
    EIter eIt = M_edgeIter(mesh);
    while ( ( edge = EIter_next(eIt) ) ) {

      if ( E_whatInType(edge) != 3 ) continue;

      pFace face = E_face(edge,0);
      pRegion start_region = F_region(face,0);
      
      pFace current_face = face;
      pRegion current_region = start_region;
      pFace next_face;
      pRegion next_region = NULL;
      while ( next_region != start_region )
      {
        next_face = E_otherFace(edge, current_face, current_region);
        next_region = F_region(next_face, 0);
        if( next_region == current_region ) {
          next_region = F_region(next_face, 1);
        }
        current_face   = next_face;
        current_region = next_region;
        
        if ( !next_region ) {
          flag = 0;
          if (verbose) {
            out << "Found a wrong edge to region connectivity !\n";
            E_info(edge,out);
          }
          break;
        }
      }
    }
    EIter_delete(eIt);
    return flag;
  }

  // -------------------------------------------------------------------
  int checkFaceToRegionConnectivity(MDB_Mesh * mesh, int verbose, std::ostream& out) {

#ifdef PARALLEL
    // In a mesh partitioned with MDB, the parallel boundary 
    // faces are not classified on a model face and the following 
    // check fails
    return 1;
#endif

    if ( M_dim(mesh) <= 2 ) return 1;

    int flag = 1;
    pFace f;
    FIter fIt= M_faceIter(mesh);
    while ( ( f = FIter_next(fIt) ) ) {
      int ok = 1;
      int nRgn = F_numRegions(f); 
      if (nRgn==0) out << "nRgn: " << nRgn << "\n";
      int type = F_whatInType(f);
      if (              nRgn == 0 ) ok = 0;
      if ( type == 2 && nRgn != 1 ) ok = 0;
      if ( type == 3 && nRgn != 2 ) ok = 0;
      if (!ok) {
        flag = 0;
        if (verbose) {
          out << "Found a wrong face to region connectivity !\n";
          F_info(f,out);
        }
        break;
      }  
    }
    FIter_delete(fIt);
    return flag;
  }

  // -------------------------------------------------------------------
  // Check that there is at least one face attached to each edge in 2D, 
  // one region and two faces in 3D.
  int checkEdgeConnectivity(MDB_Mesh * mesh, int verbose, std::ostream& out) {

#ifdef PARALLEL
    // In a mesh partitioned with MDB, the parallel boundary 
    // edges are not classified on a model line and the following 
    // check fails
    return 1;
#endif

    int dim = M_dim(mesh);
    if ( dim <= 1 ) return 1;

    int flag = 1;
    pEdge e;
    EIter eIt= M_edgeIter(mesh);
    while ( ( e = EIter_next(eIt) ) ) {
      int nFace = E_numFaces(e);
      int nRgn  = E_numRegions(e);
      if ( dim == 2 ) {
        if ( (nRgn != 0) || (nFace < 1) ) {
          flag = 0; break;
        }
      }
      else if ( dim == 3 ) {
        if ( (nRgn < 1) || (nFace < 2) ) {
          flag = 0; break;
        }
      }
    }
    EIter_delete(eIt);
    if ( (!flag) && verbose ) {
      out << "Found an ill-connected edge !\n";
      E_info(e,out);
    }
    return flag;
  }

  // -------------------------------------------------------------------
  int checkEntityPointers(MDB_Mesh * mesh, int verbose, std::ostream& out) {

    int flag = 1;

    VIter vIt = M_vertexIter(mesh);
    pVertex pv;
    while ( ( pv = VIter_next(vIt) ) ) {
      if ( !pv ) {
        flag = 0;
        if (verbose) out << "Found a null pointer when iterating on vertices\n";
      } 
    }
    VIter_delete(vIt);

    EIter eIt = M_edgeIter(mesh);
    pEdge pe;
    while ( ( pe = EIter_next(eIt) ) ) {
      if ( !pe ) {
        flag = 0;
        if (verbose) out << "Found a null pointer when iterating on edges\n";
      } 
    }
    EIter_delete(eIt);

    FIter fIt = M_faceIter(mesh);
    pFace pf;
    while ( ( pf = FIter_next(fIt) ) ) {
      if ( !pf ) {
        flag = 0;
        if (verbose) out << "Found a null pointer when iterating on faces\n";
      } 
    }
    FIter_delete(fIt);

    RIter rIt = M_regionIter(mesh);
    pRegion pr;
    while ( ( pr = RIter_next(rIt) ) ) {
      if ( !pr ) {
        flag = 0;
        if (verbose) out << "Found a null pointer when iterating on regions\n";
      } 
    }
    RIter_delete(rIt);
  
    return flag;
  }

  // -------------------------------------------------------------------
  int checkIterators(MDB_Mesh * mesh, int verbose, std::ostream& out) {

    int flag = 1;

    int nbMeshV = M_numVertices(mesh);
    int nbIterV = 0;
    VIter vIt= M_vertexIter(mesh);
    while ( VIter_next(vIt) )  nbIterV++;
    VIter_delete(vIt);
    if ( nbMeshV != nbIterV ) {
      flag = 0;
      if (verbose) {
        out << "Incoherent number of vertices: " 
            << nbMeshV << " in mesh, " 
            << nbIterV << " in iterator\n";
      }
    }

    int nbMeshE = M_numEdges(mesh);
    int nbIterE = 0;
    EIter eIt= M_edgeIter(mesh);
    while ( EIter_next(eIt) )  nbIterE++;
    EIter_delete(eIt);
    if ( nbMeshE != nbIterE ) {
      flag = 0;
      if (verbose) {
        out << "Incoherent number of edges: "
            << nbMeshE << " in mesh, "
            << nbIterE << " in iterator\n";
      }
    }

    int nbMeshF = M_numFaces(mesh);
    int nbIterF = 0;
    FIter fIt= M_faceIter(mesh);
    while ( FIter_next(fIt) )  nbIterF++;
    FIter_delete(fIt);
    if ( nbMeshF != nbIterF ) {
      flag = 0;
      if (verbose) {
        out << "Incoherent number of faces: "
            << nbMeshF << " in mesh, "
            << nbIterF << " in iterator\n";
      }
    }

    int nbMeshR = M_numRegions(mesh);
    int nbIterR = 0;
    RIter rIt= M_regionIter(mesh);
    while ( RIter_next(rIt) )  nbIterR++;
    RIter_delete(rIt);
    if ( nbMeshR != nbIterR ) {
      flag = 0;
      if (verbose) {
        out << "Incoherent number of regions: "
            << nbMeshR << " in mesh, "
            << nbIterR << " in iterator\n";
      }
    }

    return flag;
  }

  // -------------------------------------------------------------------
  int checkParameters(MDB_Mesh * mesh, int verbose, std::ostream& out)
  {
    if ( !M_isParametric(mesh) ) return 1;

    int flag = 1;

    pVertex pv;
    double tmp0,tmp1;
    VIter vIt= M_vertexIter(mesh);
    while ( ( pv = VIter_next(vIt) ) ) {
      int gDim = V_whatInType(pv);
      bool param = V_params(pv,&tmp0,&tmp1);
      if ( ( gDim==0 || gDim==1 || gDim==2 ) && !param ) {
        flag = 0;
        if (verbose) {
          out << "Non parametric point classified on geo entity with dim " <<gDim<<"\n";
          V_info(pv,out);
        }
      }
      if ( ( gDim==3 ) && param ) {
        flag = 0;
        if (verbose) {
          out << "Parametric point classified on geo entity with dim " <<gDim<<"\n";
          V_info(pv,out);
        }
      }
    }
    VIter_delete(vIt);

    return flag;
  }

  // -------------------------------------------------------------------
  bool checkMesh(MDB_Mesh * mesh, checkType type, int verbose, 
                 std::ostream& out, MeshStatus * status) {

    switch (type) {
    case CHECK_ALL: {
      if ( !checkRegionsVolume(mesh, verbose, out) ) {
        if(status) *status = NEGATIVE_VOLUME; 
        return false;
      }
      if ( !checkEdgeToRegionConnectivity(mesh, verbose, out) ) {
        if(status) *status = WRONG_EDGE_TO_RGN_CONN;
        return false;
      }
      if ( !checkFaceToRegionConnectivity(mesh, verbose, out) ) {
        if(status) *status = WRONG_FACE_TO_RGN_CONN;
        return false;
      }
      if ( !checkEdgeConnectivity(mesh, verbose, out) ) {
        if(status) *status = WRONG_EDGE_CONN;
        return false;
      }
      if ( !checkEntityPointers(mesh, verbose, out) ) {
        if(status) *status = WRONG_ENTITY_POINTERS;
        return false;
      }
      if ( !checkIterators(mesh, verbose, out) ) {
        if(status) *status = WRONG_ITERATORS;
        return false;
      }
      if ( !checkParameters(mesh, verbose, out) ) {
        if(status) *status = WRONG_PARAMETERS;
        return false;
      }
      break;
    }
    case CHECK_VOLUME: {
      if ( !checkRegionsVolume(mesh, verbose, out) ) {
        if(status) *status = NEGATIVE_VOLUME;
        return false;
      }
      break;
    }
    case CHECK_EDGE_TO_RGN_CONN: {
      if ( !checkEdgeToRegionConnectivity(mesh, verbose, out) ) {
        if(status) *status = WRONG_EDGE_TO_RGN_CONN;
        return false;
      }
      break;
    }
    case CHECK_FACE_TO_RGN_CONN: {
      if ( !checkFaceToRegionConnectivity(mesh, verbose, out) ) {
        if(status) *status = WRONG_FACE_TO_RGN_CONN;
        return false;
      }
      break;
    }
    case CHECK_EDGE_CONN: {
      if ( !checkEdgeConnectivity(mesh, verbose, out) ) {
        if(status) *status = WRONG_EDGE_CONN;
        return false;
      }
      break;
    }
    case CHECK_ENTITY_POINTERS: {
      if ( !checkEntityPointers(mesh, verbose, out) ) {
        if(status) *status = WRONG_ENTITY_POINTERS;
        return false;
      }
      break;
    }
    case CHECK_ITERATORS: {
      if ( !checkIterators(mesh, verbose, out) ) {
        if(status) *status = WRONG_ITERATORS;
        return false;
      }
      break;
    }
    case CHECK_PARAMETERS: {
      if ( !checkParameters(mesh, verbose, out) ) {
        if(status) *status = WRONG_PARAMETERS;
        return false;
      }
      break;
    }
    default: {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Not a valid checkType: %d",type);
      return false;
    }
    }

    if(status) *status = VALID;
    return true;
  }

  // -------------------------------------------------------------------

}
