// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: J.-F. Remacle, C. Dobrzynski, K. Hillewaert, G. Compere
// -------------------------------------------------------------------

#include "assert.h"
#include "MeshDataBase.h"
#include "MeshDataBaseIO.h"
#include "MSops.h"
#include "MeshDataBaseInterface.h"
#include "MeshDataBaseParallelInterface.h"
#include "MeshDataBaseParallelIO.h"
#include "MeshDataBaseMessage.h"
#include "MeshDataBaseCommCheck.h"
#include "MeshDataBaseComm.h"
#ifdef PARALLEL
#include "mpi.h"
#include "autopack.h"
#endif

#include <sstream>
using std::ostringstream;
#include <fstream>
using std::ofstream;

namespace MAd {

  // -------------------------------------------------------------------
#ifdef PARALLEL
  void M_writeParallel(pMesh m, const char * filename, int version)
  {
    SaveGmshMeshParallel(m, filename, version);
  }
#endif

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
#ifdef PARALLEL
  bool V_isInterface(pVertex pv)
  {
    pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
    void *temp_ptr;
    int isInterface = EN_getDataPtr((pEntity) pv, tagData, &temp_ptr);
    if (isInterface) return true;
    return false;
  }
#endif

  // -------------------------------------------------------------------
#ifdef PARALLEL
  int V_listInterface(pVertex pv, std::vector<int>* distProcs)
  {
    int nbDistProcs = 0;

    pMeshDataId tagVertex = MD_lookupMeshDataId("RemotePoint");

    void *temp_ptr;
    int isInterface = EN_getDataPtr((pEntity) pv, tagVertex, &temp_ptr);
    if(isInterface) {
      const std::vector<std::pair<int,pVertex> > *recup = 
        (const std::vector<std::pair<int,pVertex> > *) temp_ptr;
    
      nbDistProcs = (*recup).size();
      if (distProcs) {
        for(int j = 0 ; j<nbDistProcs ; j++) {
          distProcs->push_back( (*recup)[j].first );
        }
      }
    }
    return(nbDistProcs);
  }
#endif

  // -------------------------------------------------------------------
  /*!
    An edge is at an interface if and only if all the following conditions hold:
  
    For 2D meshes:
    - the edge has less than 2 neighbouring faces in this partition 
    and does not lie on a model edge
  
    For 3D meshes:
    - the edge has at least one neighbouring face on an interface

    \todo 
    - GCREMARK: if a model edge lies inside a surface, this condition is
    not right. Optional cross-check: its nodes are all on the same
    parallel boundary
    - GCTODO: add the case in which two geometric volumes intersect by
    an edge and the edge is on a parallel boundary
    - KHREMARK: shouldn't we reformulate this function as a check for
    potential parallel face during the establishment of connectivity,
    then implement E_isInterface with a check on the "remotePoint" data
    as for vertices. This would necessitate on the one hand all
    correspondance to be established, but would on the other hand allow
    for a real verification.
  */    
  // -------------------------------------------------------------------
#ifdef PARALLEL
  bool E_isInterface(pMesh mesh, pEdge pe)
  {
    int dim = (mesh->tets.empty()) ? 2 : 3;

    if (dim == 2) {

      // --- check that the edge is interior to a surface ---
      // --- KHREMARK: not correct 

      if ( E_whatInType(pe) == 1 ) return false;

      // --- check that the edge has less than two neighbouring faces ---
      int numf = pe->numfaces();
      if( numf == 2 ) return false;

      // --- check that the edge is not alone or
      //     with multiple neighbouring faces (debug) ---
      assert(numf == 1);

      return true;
    }
  
    else if (dim == 3) {

      bool yesItIs = false;

      // --- check that the edge has at least one neighbouring face ---
      int numf = pe->numfaces();
      for (int iF=0; iF < numf; iF++) {
        pFace pf = E_face(pe,iF);
        if ( F_isInterface(mesh, pf) ) {
          yesItIs = true;
          break;
        }
      }

      return yesItIs;
    }

    return false;
  }
#endif

  // -------------------------------------------------------------------
#ifdef PARALLEL
  bool E_isInterface(pMesh mesh, pEdge pe, int * distProc, 
                     std::vector<pVertex>* distVt)
  {
    if (distProc) *distProc = -1;
    if (distVt)   (*distVt).clear();

    if ( !E_isInterface(mesh,pe) ) return false;

    // GCREMARK: the rest of the function gives informations about the 
    //           distant partition(s) but is just a cross-check.

    // --- check that the two nodes are on a parallel boundary ---
    // KHREMARK :
    // - this is not conclusive, we should verify that the edge exists on the other partitions
    // - on the other hand, if the edge is possibly on more than two partitions,
    //   we will not flag all remote partitions
  
    pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
    pVertex pv1 = pe->p1;
    pVertex pv2 = pe->p2;
    void * temp_ptr1, *temp_ptr2;
    int isInterface1 = EN_getDataPtr((pEntity) pv1 , tagData, &temp_ptr1);
    int isInterface2 = EN_getDataPtr((pEntity) pv2 , tagData, &temp_ptr2);
    if( !(isInterface1 && isInterface2) ) {
      Msg(MDB_FATAL,"Error: found an edge which nodes are not both on a parallel boundary although it fits the conditions to be a parallel boundary edge\n");
    }

    // --- check that the two nodes are on the same parallel boundary ---
    const std::vector<std::pair<int,pVertex> > *recup1 = 
      (const std::vector<std::pair<int,pVertex> > *) temp_ptr1;
    const std::vector<std::pair<int,pVertex> > *recup2 = 
      (const std::vector<std::pair<int,pVertex> > *) temp_ptr2;
        
    std::vector<std::pair<int,pVertex> >::const_iterator rIter1 = recup1->begin();
    std::vector<std::pair<int,pVertex> >::const_iterator rIter2 = recup2->begin();
        
    for (; rIter1 != recup1->end(); rIter1++) {
      for (; rIter2 != recup2->end(); rIter2++) {
        if ( (*rIter1).first == (*rIter2).first ) {
          if (distProc) *distProc = (*rIter1).first;
          if (distVt) {
            (*distVt).clear();
            (*distVt).push_back((*rIter1).second);
            (*distVt).push_back((*rIter2).second);
          }
          return true;
        }
      }
      rIter2 = recup2->begin();
    }

    Msg(MDB_FATAL,"Error: found an edge which nodes have no common parallel boundary although it fits the conditions to be a parallel boundary edge\n");

    return false;
  }

  /*!\brief  Modified version for getting edges on multiple partitions */


  struct edge_remote_def {
    int distProc;
    pVertex vtx[2];
  };  

  bool E_isInterface(pMesh mesh, pEdge pe, std::vector<edge_remote_def>& distVt)

  {
    distVt.clear();
  
    if ( !E_isInterface(mesh,pe) ) return false;
  
    pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
    pVertex pv1 = pe->p1;
    pVertex pv2 = pe->p2;
    void * temp_ptr1, *temp_ptr2;
    int isInterface1 = EN_getDataPtr((pEntity) pv1 , tagData, &temp_ptr1);
    int isInterface2 = EN_getDataPtr((pEntity) pv2 , tagData, &temp_ptr2);
  
    if( !(isInterface1 && isInterface2) ) return false;

    // --- check that the two nodes are on the same parallel boundary ---
    const std::vector<std::pair<int,pVertex> > *recup1 = 
      (const std::vector<std::pair<int,pVertex> > *) temp_ptr1;
    const std::vector<std::pair<int,pVertex> > *recup2 = 
      (const std::vector<std::pair<int,pVertex> > *) temp_ptr2;
        
    std::vector<std::pair<int,pVertex> >::const_iterator rIter1 = recup1->begin();
    std::vector<std::pair<int,pVertex> >::const_iterator rIter2 = recup2->begin();
  
    for (; rIter1 != recup1->end(); rIter1++) {
      for (; rIter2 != recup2->end(); rIter2++) {      
        if ( (*rIter1).first == (*rIter2).first ) {
          edge_remote_def ee;
          ee.distProc = (*rIter1).first;
          ee.vtx[0] = (*rIter1).second;
          ee.vtx[1] = (*rIter2).second;
          distVt.push_back(ee);
        }
      }
      rIter2 = recup2->begin();
    }
    return (!distVt.empty());
  }
#endif

  // -------------------------------------------------------------------
  /*!
    A face is at an interface if and only if all the following conditions hold:
    - the mesh is 3D
    - the face has less than 2 neighbouring tets in this partition
    and does not lie on a model face
    
  
    \todo 
    - GCREMARK: if a model face lies inside a volume, this condition
    is not right
    - GCREMARK: this condition
    does not work with a mesh partitionned by japp (Cenaero) because
    japp flag a parallel boundary element as being part of a model
    entity with the same dimension as the element and same tag as the
    volume. Optional cross-check: its nodes should be on the same
    parallel boundary
    - KHREMARK: shouldn't we reformulate this function as a check for
    potential parallel face during the establishment of connectivity,
    then implement E_isInterface with a check on the "remotePoint" data
    as for vertices. This would necessitate on the one hand all
    correspondance to be established, but would on the other hand allow
    for a real verification.
  
  */
  // -------------------------------------------------------------------
#ifdef PARALLEL
  bool F_isInterface(pMesh mesh, pFace pface)
  {
    // --- check that the mesh is 3D ---
    if (mesh->tets.empty()) return false;
  
    // --- check that the face has less than two neighbouring regions ---
    int numtet = pface->getNbRegions();
    if (numtet == 2){
      return false;
    }
 
    // --- check that the face is interior to a volume ---
    // --- KHREMARK: this is not correct
 
    int modelDim = GEN_type(pface->g);
    if (modelDim != 3) {
      return false;
    }

    // --- check that the face is not alone (debug) ---
    assert(numtet);
    return true;
  }
#endif

  // -------------------------------------------------------------------
#ifdef PARALLEL
  // bool F_isInterface(pMesh mesh, pFace pface, int * distProc,
  //                    std::vector<pVertex>* distVt)
  // {
  //   if (distProc) *distProc = -1;
  //   if (distVt)   (*distVt).clear();

  //   if ( !F_isInterface(mesh,pface) ) return false;

  //   // GCREMARK: the rest of the function gives informations about the 
  //   //           distant partition but is just a cross-check.

  //   // --- check that the three nodes are on a parallel boundary ---
  //   pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
  //   pVertex nod[4];


  //   switch (pface->getNbNodes()) {

  //   case 3:
  //     {
  //       pface->getNodes(nod);
  //       void * temp_ptr1, *temp_ptr2, *temp_ptr3;
  //       int isInterface1 = EN_getDataPtr((pEntity) nod[0], tagData, &temp_ptr1);
  //       int isInterface2 = EN_getDataPtr((pEntity) nod[1], tagData, &temp_ptr2);
  //       int isInterface3 = EN_getDataPtr((pEntity) nod[2], tagData, &temp_ptr3);
      
      
  //       if( !(isInterface1 && isInterface2 && isInterface3) ) {
  //         Msg(MDB_FATAL,"Error: found a face which nodes are not all on a parallel boundary although it fits the conditions to be a parallel boundary face\n");
  //       }
      
  //       // --- check that the three nodes are on the same parallel boundary ---
  //       const std::vector<std::pair<int,pVertex> > *recup1 = 
  //         (const std::vector<std::pair<int,pVertex> > *) temp_ptr1;
  //       const std::vector<std::pair<int,pVertex> > *recup2 = 
  //         (const std::vector<std::pair<int,pVertex> > *) temp_ptr2;
  //       const std::vector<std::pair<int,pVertex> > *recup3 = 
  //         (const std::vector<std::pair<int,pVertex> > *) temp_ptr3;
  //       std::vector<std::pair<int,pVertex> >::const_iterator rIter1 = recup1->begin();
  //       std::vector<std::pair<int,pVertex> >::const_iterator rIter2 = recup2->begin();
  //       std::vector<std::pair<int,pVertex> >::const_iterator rIter3 = recup3->begin();
  //       for (; rIter1 != recup1->end(); rIter1++) {
  //         for (; rIter2 != recup2->end(); rIter2++) {
  //           for (; rIter3 != recup3->end(); rIter3++) {
  //             if ( (*rIter1).first == (*rIter2).first &&
  //                  (*rIter2).first == (*rIter3).first ) {
  //               if (distProc) *distProc = (*rIter1).first;
  //               if (distVt) {
  //                 (*distVt).push_back((*rIter1).second);
  //                 (*distVt).push_back((*rIter2).second);
  //                 (*distVt).push_back((*rIter3).second);
  //               }
  //               return true;
  //             }
  //           }
  //           rIter3 = recup3->begin();
  //         }
  //         rIter2 = recup2->begin();
  //       }
  //       break;
  //     }
    
  //   case 4:
  //     {
  //       pface->getNodes(nod);
  //       void * temp_ptr1, *temp_ptr2, *temp_ptr3, *temp_ptr4;
  //       int isInterface1 = EN_getDataPtr((pEntity) nod[0], tagData, &temp_ptr1);
  //       int isInterface2 = EN_getDataPtr((pEntity) nod[1], tagData, &temp_ptr2);
  //       int isInterface3 = EN_getDataPtr((pEntity) nod[2], tagData, &temp_ptr3);
  //       int isInterface4 = EN_getDataPtr((pEntity) nod[3], tagData, &temp_ptr4);
      
    
  //       if( !(isInterface1 && isInterface2 && isInterface3 && isInterface4) ) {
  //         Msg(MDB_FATAL,"Error: found a face which nodes are not all on a parallel boundary although it fits the conditions to be a parallel boundary face\n");
  //       }
    
  //       // --- check that the three nodes are on the same parallel boundary ---
  //       const std::vector<std::pair<int,pVertex> > *recup1 = 
  //         (const std::vector<std::pair<int,pVertex> > *) temp_ptr1;
  //       const std::vector<std::pair<int,pVertex> > *recup2 = 
  //         (const std::vector<std::pair<int,pVertex> > *) temp_ptr2;
  //       const std::vector<std::pair<int,pVertex> > *recup3 = 
  //         (const std::vector<std::pair<int,pVertex> > *) temp_ptr3;
  //       const std::vector<std::pair<int,pVertex> > *recup4 = 
  //         (const std::vector<std::pair<int,pVertex> > *) temp_ptr4;
  //       std::vector<std::pair<int,pVertex> >::const_iterator rIter1 = recup1->begin();
  //       std::vector<std::pair<int,pVertex> >::const_iterator rIter2 = recup2->begin();
  //       std::vector<std::pair<int,pVertex> >::const_iterator rIter3 = recup3->begin();
  //       std::vector<std::pair<int,pVertex> >::const_iterator rIter4 = recup4->begin();
  //       for (; rIter1 != recup1->end(); rIter1++) {
  //         for (; rIter2 != recup2->end(); rIter2++) {
  //           for (; rIter3 != recup3->end(); rIter3++) {
  //             for (; rIter4 != recup4->end(); rIter4++) {
  //               if ( (*rIter1).first == (*rIter2).first &&
  //                    (*rIter2).first == (*rIter3).first &&
  //                    (*rIter3).first == (*rIter4).first ) {
  //                 if (distProc) *distProc = (*rIter4).first;
  //                 if (distVt) {
  //                   (*distVt).push_back((*rIter1).second);
  //                   (*distVt).push_back((*rIter2).second);
  //                   (*distVt).push_back((*rIter3).second);
  //                   (*distVt).push_back((*rIter4).second);
  //                 }
  //                 return true;
  //               }
  //             }
  //             rIter4 = recup4->begin();
  //           }
  //           rIter3 = recup3->begin();
  //         }
  //         rIter2 = recup2->begin();
  //       }
  //       break;
  //     }
  //   }
    
  //   Msg(MDB_FATAL,"Error: found a face which nodes have no common parallel boundary although it fits the conditions to be a parallel boundary face\n");

  //   return false;
  // }

  /*! \brief corrected version of F_isInterface for the situation in which another proc holds all of the nodes
    of the face, without containing the actual face ... - yes, shit happens ... */ 

  struct face_comm
  {
    int     nbNodes;
    pVertex p1,p2,p3,p4;   // pointer in dest
    int     sendID;       
  };

  bool F_isInterface(pMesh mesh, pFace pface, std::vector<face_comm>& distFace)
  {

    if ( !F_isInterface(mesh,pface) ) return false;

    // GCREMARK: the rest of the function gives informations about the 
    //           distant partition but is just a cross-check.

    // --- check that the three nodes are on a parallel boundary ---
    pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
    pVertex nod[4];

    distFace.clear();


    switch (pface->getNbNodes()) {

    case 3:
      {
        pface->getNodes(nod);
        void * temp_ptr1, *temp_ptr2, *temp_ptr3;
        int isInterface1 = EN_getDataPtr((pEntity) nod[0], tagData, &temp_ptr1);
        int isInterface2 = EN_getDataPtr((pEntity) nod[1], tagData, &temp_ptr2);
        int isInterface3 = EN_getDataPtr((pEntity) nod[2], tagData, &temp_ptr3);
      
      
        if( !(isInterface1 && isInterface2 && isInterface3) ) {
          return false;
          //Msg(MDB_FATAL,"Error: found a face which nodes are not all on a parallel boundary although it fits the conditions to be a parallel boundary face\n");
        }
      
        // --- check that the three nodes are on the same parallel boundary ---
        const std::vector<std::pair<int,pVertex> > *recup1 = 
          (const std::vector<std::pair<int,pVertex> > *) temp_ptr1;
        const std::vector<std::pair<int,pVertex> > *recup2 = 
          (const std::vector<std::pair<int,pVertex> > *) temp_ptr2;
        const std::vector<std::pair<int,pVertex> > *recup3 = 
          (const std::vector<std::pair<int,pVertex> > *) temp_ptr3;
        std::vector<std::pair<int,pVertex> >::const_iterator rIter1 = recup1->begin();
        std::vector<std::pair<int,pVertex> >::const_iterator rIter2 = recup2->begin();
        std::vector<std::pair<int,pVertex> >::const_iterator rIter3 = recup3->begin();
        for (; rIter1 != recup1->end(); rIter1++) {
          for (; rIter2 != recup2->end(); rIter2++) {
            for (; rIter3 != recup3->end(); rIter3++) {
              if ( (*rIter1).first == (*rIter2).first &&
                   (*rIter2).first == (*rIter3).first ) {

                face_comm fc;
              
                fc.sendID = (*rIter1).first;
                fc.nbNodes = 3;

                fc.p1 = (*rIter1).second;
                fc.p2 = (*rIter2).second;
                fc.p3 = (*rIter3).second;
              
                distFace.push_back(fc);
              }
            }
            rIter3 = recup3->begin();
          }
          rIter2 = recup2->begin();
        }
        break;
      }
    
    case 4:
      {
        pface->getNodes(nod);
        void * temp_ptr1, *temp_ptr2, *temp_ptr3, *temp_ptr4;
        int isInterface1 = EN_getDataPtr((pEntity) nod[0], tagData, &temp_ptr1);
        int isInterface2 = EN_getDataPtr((pEntity) nod[1], tagData, &temp_ptr2);
        int isInterface3 = EN_getDataPtr((pEntity) nod[2], tagData, &temp_ptr3);
        int isInterface4 = EN_getDataPtr((pEntity) nod[3], tagData, &temp_ptr4);
      
    
        if( !(isInterface1 && isInterface2 && isInterface3 && isInterface4) ) {
          Msg(MDB_FATAL,"Error: found a face which nodes are not all on a parallel boundary although it fits the conditions to be a parallel boundary face\n");
        }
    
        // --- check that the three nodes are on the same parallel boundary ---
        const std::vector<std::pair<int,pVertex> > *recup1 = 
          (const std::vector<std::pair<int,pVertex> > *) temp_ptr1;
        const std::vector<std::pair<int,pVertex> > *recup2 = 
          (const std::vector<std::pair<int,pVertex> > *) temp_ptr2;
        const std::vector<std::pair<int,pVertex> > *recup3 = 
          (const std::vector<std::pair<int,pVertex> > *) temp_ptr3;
        const std::vector<std::pair<int,pVertex> > *recup4 = 
          (const std::vector<std::pair<int,pVertex> > *) temp_ptr4;
        std::vector<std::pair<int,pVertex> >::const_iterator rIter1 = recup1->begin();
        std::vector<std::pair<int,pVertex> >::const_iterator rIter2 = recup2->begin();
        std::vector<std::pair<int,pVertex> >::const_iterator rIter3 = recup3->begin();
        std::vector<std::pair<int,pVertex> >::const_iterator rIter4 = recup4->begin();
        for (; rIter1 != recup1->end(); rIter1++) {
          for (; rIter2 != recup2->end(); rIter2++) {
            for (; rIter3 != recup3->end(); rIter3++) {
              for (; rIter4 != recup4->end(); rIter4++) {
                if ( (*rIter1).first == (*rIter2).first &&
                     (*rIter2).first == (*rIter3).first &&
                     (*rIter3).first == (*rIter4).first ) {

                
                  face_comm fc;
                
                  fc.sendID = (*rIter1).first;

                  fc.nbNodes = 4;

                  fc.p1 = (*rIter1).second;
                  fc.p2 = (*rIter1).second;
                  fc.p3 = (*rIter2).second;
                  fc.p4 = (*rIter3).second;
                
                  distFace.push_back(fc);
                
                }
              }
              rIter4 = recup4->begin();
            }
            rIter3 = recup3->begin();
          }
          rIter2 = recup2->begin();
        }
        break;
      }
    }
    
    // Msg(MDB_FATAL,"Error: found a face which nodes have no common parallel boundary although it fits the conditions to be a parallel boundary face\n");

    return (distFace.size() != 0);
  }
#endif

  // -------------------------------------------------------------------
#ifdef PARALLEL
  bool R_isInterface(pRegion pr) {
    return false;
  }
#endif

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
#ifdef PARALLEL
  void createRemotePointLists (pMesh mesh, pMeshDataId tagVertex)
  {
    int myrank, nbproc;
    MPI_Comm_size(MPI_COMM_WORLD, &nbproc);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    // --- list local nodes on parallel boundaries ---
    std::set<std::pair<int,pVertex> > bdryNodes;
  
    int meshDim = (mesh->tets.empty()) ? 2 : 3;
    if ( meshDim == 3 ) {
      FIter fit = M_faceIter(mesh);
      while ( pFace pf = FIter_next(fit) ) {
        if ( F_isInterface(mesh,pf) ) {
          for (int iv = 0; iv < F_numVertices(pf); iv++) {
            pVertex pv = F_vertex(pf,iv);
            bdryNodes.insert(std::make_pair(pv->iD,pv));
          }
          // GCTODO: add high order points (check if necessary first)
          // KHREMARK: higher order points not needed 
        }
      }
    } else {
      EIter eit = M_edgeIter(mesh);
      while ( pEdge pe = EIter_next(eit) ) {
        if ( E_isInterface(mesh,pe) ) {
          for (int iv = 0; iv < 2; iv++) {
            pVertex pv = E_vertex(pe,iv);
            bdryNodes.insert(std::make_pair(pv->iD,pv));
          }
          // GCTODO: check if high order points are necessary
          // KHREMARK: not necessary 
          for (int ip = 0; ip < E_numPoints(pe); ip++) {
            pPoint pp = E_point(pe,ip);
            bdryNodes.insert(std::make_pair(pp->iD,(pVertex)pp));
          }
        }
      }
    }

    MPI_Barrier(MPI_COMM_WORLD);

    // --- send list ---

    int* sendcounts = new int[nbproc];

    for (int iProcDest=0; iProcDest < nbproc; iProcDest++) {
      if ( iProcDest == myrank ) {
        sendcounts[iProcDest] = 0;
        continue;    
      }
      sendcounts[iProcDest] = 1;
      int size = bdryNodes.size()*sizeof(int);
      void *buf = AP_alloc(iProcDest,669,size);
      int * bufCast = reinterpret_cast<int *>(buf);
      std::set<std::pair<int,pVertex> >::const_iterator nIter = bdryNodes.begin();
      std::set<std::pair<int,pVertex> >::const_iterator nLast = bdryNodes.end();
      for (; nIter != nLast; nIter++)   *(bufCast++) = (*nIter).first;
      AP_send(buf);
    }

    AP_check_sends(AP_NOFLAGS);
    AP_reduce_nsends(sendcounts);
    AP_flush();

    // --- receive lists ---

    int message = 0;
    int count;
    while (!AP_recv_count(&count) || message < count) {

      void *msg;
      int from;
      int tag;
      int size;
      int rc;
    
      rc = AP_recv (MPI_ANY_SOURCE, 669, AP_BLOCKING|AP_DROPOUT,
                    &msg, &size, &from, &tag);
    
      if (rc) {
      
        int* msgCast = reinterpret_cast<int*>(msg);
      
        std::set<int> distNodes;
        int nbDistNodes = size / sizeof(int);
        for (int iDistN = 0; iDistN < nbDistNodes; iDistN++) {
          distNodes.insert(msgCast[iDistN]);
        }

        std::set<std::pair<int,pVertex> >::const_iterator nIter = bdryNodes.begin();
        std::set<std::pair<int,pVertex> >::const_iterator nLast = bdryNodes.end();
        for (; nIter != nLast; nIter++) {
          if ( distNodes.find( (*nIter).first ) != distNodes.end() ) {
            pVertex pv = (*nIter).second;
            void *temp_ptr;
            int exist = EN_getDataPtr((pEntity) pv, tagVertex, &temp_ptr);
            std::vector<std::pair<int,pVertex> > * parts;
            if (exist) {
              parts = reinterpret_cast<std::vector<std::pair<int,pVertex> >*>(temp_ptr);
            } else {
              parts = new std::vector<std::pair<int,pVertex> >;
              EN_attachDataPtr((pEntity) pv, tagVertex, parts);
            }
            parts->push_back(std::pair <int, MDB_Point*>(from,NULL));
          }
        }

        message++;
        AP_free(msg);
      }
    }

    // --- wait until all have finished sending ---
  
    AP_check_sends(AP_WAITALL);
    delete [] sendcounts;

    // GCDEBUG
    //   MPI_Barrier(MPI_COMM_WORLD);
  }
#endif

  // -------------------------------------------------------------------
#ifdef PARALLEL
  struct point_comm
  {
    pVertex p;     // local pointer
    int nID;       // myrank
    int numGlobal; // Id global for the vertex
  };
#endif

  // -------------------------------------------------------------------
#ifdef PARALLEL
  void V_createInfoInterface(pMesh mesh, pMeshDataId tagVertex)
  {
    int mysize,myrank;
    MPI_Comm_size(MPI_COMM_WORLD, &mysize);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    // --- Delete previous information ---
    VIter vit = M_vertexIter(mesh);
    while ( pVertex pv = VIter_next(vit) ) {
      void * tmp_ptr;
      if( EN_getDataPtr((pEntity) pv, tagVertex, &tmp_ptr) ) {
        EN_deleteData((pEntity) pv, tagVertex);
      }
    }
    VIter_delete(vit);

    // --- Create lists of distant procs ---
    createRemotePointLists(mesh, tagVertex);

    // --- Send node pointers ---
    int *sendcounts = new int[mysize];
    for(int i=0; i<mysize; i++) sendcounts[i] = 0;
  
    vit = M_vertexIter(mesh);
    while ( pVertex pv = VIter_next(vit) ) {
      std::vector<int> distProcTab;
      int nbDistProc = V_listInterface(pv, &distProcTab);
      if ( nbDistProc ) {
        for (int j=0; j<nbDistProc ; j++) {
          int distProc = distProcTab[j];
          assert ( distProc != myrank );
          void *buf = AP_alloc(distProc,445,sizeof(point_comm));
          point_comm *castbuf = (point_comm *) buf;
          castbuf->p         = pv;
          castbuf->nID       = myrank;
          castbuf->numGlobal = EN_id((pEntity)pv);
          AP_send(buf);
          sendcounts[distProc]++;
        }
      }
    }
    VIter_delete(vit);

    AP_check_sends(AP_NOFLAGS);
    AP_reduce_nsends(sendcounts);
    AP_flush();

    // --- Receive node pointers ---
    int message=0, count;
    while (!AP_recv_count(&count) || message<count) {
      void *msg;
      int  from;
      int  tag;
      int  size;
      int  recv;
      recv = AP_recv(MPI_ANY_SOURCE, 445, AP_BLOCKING|AP_DROPOUT,
                     &msg, &size, &from, &tag);
      if(recv) {
        message++;
        point_comm * castbuf = (point_comm*) msg;
        pVertex precv (0);
        precv = castbuf -> p;
        assert(precv);
        int recvID = castbuf -> nID;
        int numrecv = castbuf -> numGlobal;
      
        pVertex p = mesh->find_point(numrecv);
        if(p) {
          void *temp_ptr; 
          int isInterface = EN_getDataPtr((pEntity) p, tagVertex, &temp_ptr);
          assert(isInterface);
          std::vector<std::pair<int , MDB_Point*> > *recup = (std::vector<std::pair<int , MDB_Point*> > *) temp_ptr;
          for(unsigned int j=0 ; j<(*recup).size() ; j++) {
            int remoteID = (*recup)[j].first;
            if(remoteID == recvID) {
              (*recup)[j].second = precv;
            }
          }
        }
        AP_free(msg);
      }
    }
    AP_check_sends(AP_WAITALL);
    MPI_Barrier(MPI_COMM_WORLD);
    delete [] sendcounts;
  
    MDB_CommCheck cc;
    exchangeDataOnVertices(mesh,cc);
  
  }
#endif

  // -------------------------------------------------------------------
#ifdef PARALLEL
  void MDB_Mesh::bdryLinkRemotePoint()
  {
    int mysize,myrank;
    MPI_Comm_size(MPI_COMM_WORLD, &mysize);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);   

    /*send pointeurs*/
    int *sendcounts = new int[mysize];
    for(int i=0;i<mysize;i++)sendcounts[i]=0;
  
    pMeshDataId tagData = MD_lookupMeshDataId("RemotePoint");
    VIter vit = M_vertexIter(this);
    pVertex pv;  
    int nsend=0;
    while ((pv = VIter_next(vit))) {
      void *temp_ptr; 
      int isInterface = EN_getDataPtr((pEntity) pv , tagData, &temp_ptr);
    
      if(isInterface) {
        std::vector<std::pair<int , MDB_Point*> > *recup = (std::vector<std::pair<int , MDB_Point*> > *) temp_ptr;
        int numGlobal = EN_id((pEntity)pv);
        for(unsigned int j=0 ; j<(*recup).size() ; j++) {
          nsend++;
          int remoteID  = (*recup)[j].first;
          assert(remoteID != myrank);
          void *buf = AP_alloc(remoteID,444,sizeof(point_comm));
          point_comm *castbuf = (point_comm *) buf;
          castbuf->p         = pv;
          castbuf->nID       = myrank;
          castbuf->numGlobal = numGlobal;
          AP_send(buf);
          sendcounts[remoteID]++;   
        }
      }
    }	
    VIter_delete(vit);   

    AP_flush();
    AP_check_sends(AP_NOFLAGS);
    // AP_reduce_nsends(sendcounts);

    /*receive pointers*/
    int message=0;
    // while (!AP_recv_count(&count) || message<count) {
    while (message<nsend) {
      void *msg;
      int  from;
      int  tag;
      int  size;
      int  recv;
      recv = AP_recv(MPI_ANY_SOURCE, 444, AP_BLOCKING|AP_DROPOUT,
                     &msg, &size, &from, &tag);
      if(recv) {
        message++;
        point_comm * castbuf = (point_comm*) msg;
        pVertex precv (0);
        precv = castbuf -> p;
        assert(precv);
        int recvID = castbuf -> nID;
        int numrecv = castbuf -> numGlobal;
      
        pVertex p = this->find_point(numrecv);
        if(p) {
          void *temp_ptr; 
          int isInterface = EN_getDataPtr((pEntity) p , tagData, &temp_ptr);
          assert(isInterface);
          std::vector<std::pair<int , MDB_Point*> > *recup = 
            (std::vector<std::pair<int , MDB_Point*> > *) temp_ptr;
          for(unsigned int j=0 ; j<(*recup).size() ; j++) {
            int remoteID  = (*recup)[j].first;
            if(remoteID == recvID) {
              (*recup)[j].second = precv;
            }
          }  
        }
        AP_free(msg);
      }		   
   
    }
    AP_check_sends(AP_WAITALL);
    MPI_Barrier(MPI_COMM_WORLD);
    delete [] sendcounts;
  }
#endif


  // -------------------------------------------------------------------
#ifdef PARALLEL
  struct edge_comm {
    pVertex p1,p2; // pointer in dest
    pEdge pe; // pointer to the local edge
  };
#endif

  // -------------------------------------------------------------------
  /*! \brief Create edge correspondance and assure coherent orientation \ingroup parallel */ 
#ifdef PARALLEL
  void E_createInfoInterface(pMesh mesh, pMeshDataId tagEdge)
  { 
    int nproc,myrank;
    MPI_Comm_size(MPI_COMM_WORLD, &nproc);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    // --- Delete previous information ---
    EIter eit = M_edgeIter(mesh);
    while ( pEdge pe = EIter_next(eit) ) {
      void * tmp_ptr;
      if( EN_getDataPtr((pEntity) pe, tagEdge, &tmp_ptr) ) {
        EN_deleteData((pEntity) pe, tagEdge);
      }
    }
    EIter_delete(eit);

    // --- Send edge info ---
    int *sendcounts = new int[nproc];
    for(int i=0;i<nproc;i++)sendcounts[i]=0;

    eit = M_edgeIter(mesh);
    while ( pEdge pe = EIter_next(eit) ) {

      std::vector<edge_remote_def> remote;

      if (E_isInterface(mesh,pe,remote)) {

        std::vector<edge_remote_def >::const_iterator riter = remote.begin();

        for (;riter!=remote.end();++riter) {
        
          int distProc = riter->distProc;
          void *buf = AP_alloc(distProc,446,sizeof(edge_comm));
          edge_comm *castbuf = (edge_comm *) buf;
          castbuf->p1 = riter->vtx[0];
          castbuf->p2 = riter->vtx[1];
          castbuf->pe = pe;
          AP_send(buf);
          sendcounts[distProc]++;
        }
      }
    }
    EIter_delete(eit);

    AP_check_sends(AP_NOFLAGS);
    AP_reduce_nsends(sendcounts);
    //AP_flush();

    // --- Receive edge info ---
    int message=0,count;
    while (!AP_recv_count(&count) ||message<count) {
      void *msg;
      int from;
      int tag;
      int size;
      int rc;
      rc=AP_recv(MPI_ANY_SOURCE,446, AP_BLOCKING|AP_DROPOUT,
                 &msg, &size, &from, &tag);
      if (rc) {
        message++;
        edge_comm * comm = (edge_comm*) msg;
        pEdge pe = E_exist(comm->p1,comm->p2);
        if(pe) {
        
          std::map<int,pEdge>* list;
          // std::vector<std::pair<int, pEdge> >* list;
          if(EN_getDataPtr((pEntity) pe, tagEdge, (void**)&list)) {
            // list->push_back(std::pair<int, pEdge>(from,comm->pe));
            if ((from < list->begin()->first) && (from < myrank)) {
              E_align(pe,comm->p1,comm->p2);
              list->insert(std::pair<int,pEdge>(from,comm->pe));
            }
          }
          else {
            list = new std::map<int,pEdge>;
            // list = new std::vector<std::pair<int, pEdge> >;
            // list->push_back(std::pair<int, pEdge>(from,comm->pe));
            list->insert(std::pair<int,pEdge>(from,comm->pe));
            if (from < myrank) E_align(pe,comm->p1,comm->p2);
            EN_attachDataPtr((pEntity) pe, tagEdge, (void*)list);
          }
        }
        AP_free(msg);
      }
    }
    AP_check_sends(AP_WAITALL);
    MPI_Barrier(MPI_COMM_WORLD);

    delete [] sendcounts;

    MDB_CommCheck cc;
    exchangeDataOnEdges(mesh,cc);
  }
#endif

  // -------------------------------------------------------------------
  // #ifdef PARALLEL
  // struct face_comm
  // {
  //   int     nbNodes;
  //   pVertex p1,p2,p3,p4;   // pointer in dest
  //   int     sendID;       
  // }
  // #endif

  // -------------------------------------------------------------------
  /*! \brief Establish face to face correspondance and assure coherent orientation \ingroup parallel */ 
  //

#ifdef PARALLEL
  void F_createInfoInterface(pMesh mesh, pMeshDataId tagFace)
  {
    int nproc,myrank;
    MPI_Comm_size(MPI_COMM_WORLD, &nproc);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    // --- Delete previous information ---
    FIter fit = M_faceIter(mesh);
    while ( pFace pf = FIter_next(fit) ) {
      void * tmp_ptr;
      if( EN_getDataPtr((pEntity) pf, tagFace, &tmp_ptr) ) {
        EN_deleteData((pEntity) pf, tagFace);
      }
    }
    FIter_delete(fit);

    // --- Send face info ---
    int *sendcounts = new int[nproc];
    for(int i=0;i<nproc;i++)sendcounts[i]=0;
  
    fit = M_faceIter(mesh);
    while ( pFace pf = FIter_next(fit) ) {
      int distProc = -1;
      std::vector<pVertex> distVt;

      std::vector<face_comm> rd;

      if ( F_isInterface(mesh, pf, rd) ) {

        std::vector<face_comm>::iterator rIter = rd.begin();

        for (;rIter!=rd.end();++rIter) {

          distProc = rIter->sendID;

          void *buf = AP_alloc(distProc,444,sizeof(face_comm));
          face_comm *castbuf = (face_comm *) buf;
          memcpy(buf,&(*rIter),sizeof(face_comm));
          castbuf->sendID = myrank;
        
          AP_send(buf);
          sendcounts[distProc]++;
        }
      }
    }
    FIter_delete(fit);

    AP_check_sends(AP_NOFLAGS);
    AP_reduce_nsends(sendcounts);
    //AP_flush();
  
    // --- Receive face info ---
    int message=0, count;
    while (!AP_recv_count(&count) || message<count) {
      void *msg;
      int from;
      int tag;
      int size;
      int rc;
      rc=AP_recv(MPI_ANY_SOURCE,444, AP_BLOCKING|AP_DROPOUT,
                 &msg, &size, &from, &tag);
      if (rc) {
        message++;
        face_comm * castbuf = (face_comm*) msg;
        pVertex p1recv,p2recv,p3recv,p4recv;
        int nbNodes = castbuf->nbNodes;
        p1recv = castbuf -> p1;
        p2recv = castbuf -> p2;
        p3recv = castbuf -> p3;
      
        assert(p1recv);assert(p2recv);assert(p3recv);
        p4recv = nbNodes == 4 ? castbuf -> p4 : NULL;
      
        int nprocrecv = castbuf->sendID;
        assert(nprocrecv==from);
      
        // pFace pface = F_exist(2,p1recv,p2recv,p3recv,p4recv);
        pFace pface = F_exist(p1recv,p2recv,p3recv,p4recv);
        
        //KH: how can a face have more than two partitions on which it lives ?
      
        if (pface) {

          // lowest ranking processor is master
          if (from < myrank) F_align(pface,p1recv,p2recv,p3recv,p4recv);
        
          void* tmpptr;
          int is = EN_getDataPtr((pEntity) pface , tagFace,&tmpptr);
          if(is){
            std::vector<int> *recup = (std::vector<int> *) tmpptr;
            for(unsigned int i=0 ; i<(*recup).size() ; i++) assert((*recup)[i]!=nprocrecv);
            (*recup).push_back(nprocrecv);
          } else {
            std::vector<int> *list = new std::vector<int>;
            list->push_back(nprocrecv);
            EN_attachDataPtr((pEntity) pface, tagFace, list);
          }
        }
        AP_free(msg);
      }
    }
    AP_check_sends(AP_WAITALL);
    MPI_Barrier(MPI_COMM_WORLD);

    delete [] sendcounts;
  
    MDB_CommCheck cc;
    exchangeDataOnFaces(mesh,cc);
  }
#endif

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  // #ifdef PARALLEL
  // void UpdateIDGlobal(pMesh mesh, int IdGlobal)
  // {
  //   int mysize,myrank;
  //   MPI_Comm_size(MPI_COMM_WORLD, &mysize);
  //   MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

  //   // send pointers
  //   int *sendcounts = new int[mysize];
  //   for(int i=0; i<mysize; i++) sendcounts[i] = 0;
  
  //   pMeshDataId tagVertex = MD_lookupMeshDataId("RemotePoint");
  //   pMeshDataId tagGlob = MD_lookupMeshDataId("IdGlobal");

  //   VIter vit = M_vertexIter(mesh);
  //   pVertex pv;
  //   while ((pv = VIter_next(vit))) {
  //     void *temp_ptr; 
  //     int isInterface = EN_getDataPtr((pEntity) pv , tagVertex, &temp_ptr);
    
  //     if(isInterface) {
  //       std::vector<std::pair<int , MDB_Point*> > *recup = (std::vector<std::pair<int , MDB_Point*> > *) temp_ptr;
  //       int minproc = mysize + 10;
  //       for(unsigned int j=0 ; j<(*recup).size() ; j++) {
  //         minproc = std::min(minproc,(*recup)[j].first);
  //       }
  //       minproc = std::min(myrank,minproc);
  //       if(minproc!=myrank) continue;
  //       int numGlobal = EN_id((pEntity) pv) + IdGlobal + 1;
  //       EN_attachDataInt((pEntity) pv, tagGlob, numGlobal);
  //       for(unsigned int j=0 ; j<(*recup).size() ; j++) {
  //         int remoteID  = (*recup)[j].first;
  //         assert(remoteID != myrank);
  //         void *buf = AP_alloc(remoteID,444,sizeof(point_comm));
  //         point_comm *castbuf = (point_comm *) buf;
  //         castbuf->p         = (*recup)[j].second;
  //         castbuf->nID       =  myrank;
  //         castbuf->numGlobal =  numGlobal;
  //         AP_send(buf);
  //         sendcounts[remoteID]++;
  //       }
  //     }
  //   }
  //   VIter_delete(vit);

  //   AP_check_sends(AP_NOFLAGS);
  //   AP_reduce_nsends(sendcounts);
  //   AP_flush();

  //   // receive pointers
  //   int message=0;
  //   int count;
  //   while (!AP_recv_count(&count) || message<count) {
  //     void *msg;
  //     int  from;
  //     int  tag;
  //     int  size;
  //     int  recv;
  //     recv = AP_recv(MPI_ANY_SOURCE, 444, AP_BLOCKING|AP_DROPOUT,
  // 		   &msg, &size, &from, &tag);
  //     if(recv) {
  //       message++;
  //       point_comm * castbuf = (point_comm*) msg;
  //       pVertex precv (0);
  //       precv = castbuf -> p;
  //       assert(precv);
  //       //int recvID = castbuf -> nID;
  //       int numrecv = castbuf -> numGlobal;
      
  //       // DEBUG
  //       void *temp_ptr; 
  //       int isInterface = EN_getDataPtr((pEntity) precv, tagVertex, &temp_ptr);
  //       assert(isInterface);
      
  //       EN_attachDataInt((pEntity) precv, tagGlob, numrecv);        
  //       AP_free(msg);
  //     }
    
  //   }
  //   AP_check_sends(AP_WAITALL);
  //   MPI_Barrier(MPI_COMM_WORLD);
  //   delete [] sendcounts;
  // }
  // #endif

  // -------------------------------------------------------------------

}
